(**************************************************************************
* BlockCAD, a program for building with Lego like blocks on the computer
* Copyright (C) 1998  Anders Isaksson
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
* -------------------------------------------------------------------------
* I hope You send me a copy of any important changes/additions You do.
*
* E-mail: isaksson.etuna@ebox.tninet.se
*
**************************************************************************)

unit Lcadpc2;
{
  The main unit for hardcoded blocks - bricks, slopes, rounds
}

interface

uses
  Windows, Graphics, Classes,
  DrawObject
  ;

const
  PFACTORXY         = 20;
  PFACTORZ          = 24;

type

  TRGBQuadArray = array[0..MAXWORD] of TRGBQuad;
  PRGBQuadArray = ^TRGBQuadArray;

  ZPLANE = record
    A, B, C, D: single;                 //integer;
  end;

  TMyFloat = single;

  TPoint3D = record
    x, y, z: integer;
  end;

  TDPoint = record
    x, y: TMyFloat;
  end;

  TBlockClass = class of TBlock;

  TBlock = class(TObject)               { This is the base class for all blocks }
  protected
    TopCover, BotCover: integer;
    gpPointer: Pointer;                 { General purpose pointer }
    procedure Octant0(const DrawCanvas: TDrawObject); virtual;
    procedure Octant1(const DrawCanvas: TDrawObject); virtual;
    procedure Octant2(const DrawCanvas: TDrawObject); virtual;
    procedure Octant3(const DrawCanvas: TDrawObject); virtual;
    procedure Octant4(const DrawCanvas: TDrawObject); virtual;
    procedure Octant5(const DrawCanvas: TDrawObject); virtual;
    procedure Octant6(const DrawCanvas: TDrawObject); virtual;
    procedure Octant7(const DrawCanvas: TDrawObject); virtual;

    procedure TopStuds(const DrawCanvas: TDrawObject); virtual;
    procedure BottomStuds(const DrawCanvas: TDrawObject); virtual;
    function HitTest(const xx, yy: integer): Boolean; virtual;
    procedure Initialize(const FileName, PartName: string); virtual;
  public
    Selected: boolean;
    px, py, pz, pl, pb, ph: integer;
    Dir: integer;
    st: Boolean;
    cl: TColor;
    Transp: Boolean;
    bp: Boolean;
    t: Char;
    PartId: string;
    Group: integer;
    Step: integer;
    LDrawPatch: integer;
    constructor CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
      const cc: TColor;
      const ss: Boolean;
      const IsBasePlate: Boolean;
      const IsTransp: Boolean;
      const LDrawId: string); virtual;
    procedure Turn; virtual;
    procedure Draw(const DrawCanvas: TDrawObject); virtual;
    function AsString: string; virtual;
    function ZeroPoint: TPoint3D; virtual;
    function ZLevel(const xx, yy, ll, bb: integer): integer; virtual;
    class function CreateCopy(const AnotherBlock: TBlock): TBlock;
    class function GetStudQuality: integer;
    class procedure SetStudQuality(const q: integer);
    class function CreateFromString(const s: string): TBlock; virtual;
  end;

  TBlockList = class(TList)
  private
    fUndoList: TStringList;
    fRedoList: TStringList;
    fSettingsRotXPos: integer;
    fSettingsRotZPos: integer;
    fSettingsScaleBarPos: integer;
    fSettingsOffsetXPos: integer;
    fSettingsOffsetYPos: integer;
    fSettingsStuds: integer;
    fSettingsVersion: integer;
    function LocateAsString(const s: string): integer;
  public
    oct: integer;
    Step: integer;
    bx, by, bz, bl, bb, bh: integer;
    constructor Create;
    destructor Destroy; override;
    procedure Clear; override;
    procedure ClearKeepUndo;
    function Cover(const i, j: TBlock): Boolean;
    procedure Add(const Piece: TObject);
    procedure Concat(const Macro: TBlockList; const xx, yy, zz: integer);
    procedure SortForOctant(const AnOctant: integer);
    function FindBlockAtPos(const xx, yy, zz: integer): integer;
    function FindBlockAtScreenPos(const xx, yy: integer): integer;
    procedure Draw(const Canvas: TDrawObject; DrawAlways: Boolean);
    procedure UndoSave(Cmd: string; Piece: TObject);
    procedure Undo;
    procedure Redo;
    function UndoCount: integer;
    function RedoCount: integer;
    procedure CheckTopAndBottom;
    procedure LoadFromFile(FileName: string);
    property sRotXP: integer read fSettingsRotXPos;
    property sRotZP: integer read fSettingsRotZPos;
    property sScaleBP: integer read fSettingsScaleBarPos;
    property sOffsXP: integer read fSettingsOffsetXPos;
    property sOffsYP: integer read fSettingsOffsetYPos;
    property sStuds: integer read fSettingsStuds;
    property sVersion: integer read fSettingsVersion;
  end;

  TRoof = class(TBlock)
  public
    constructor CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
      const cc: TColor;
      const ss: Boolean;
      const IsBasePlate: Boolean;
      const IsTransp: Boolean;
      const LDrawId: string); override;
    procedure Draw(const DrawCanvas: TDrawObject); override;
    procedure TopStuds(const DrawCanvas: TDrawObject); override;
    function HitTest(const xx, yy: integer): Boolean; override;

  protected
    procedure xFront(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); virtual;
    procedure xBack(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); virtual;
    procedure xTop(const x, y, z, l, b, h: integer; const st: Boolean; const DrawCanvas: TDrawObject); virtual;
    procedure xBottom(const x, y, z, l, b, h: integer; const st: Boolean; const DrawCanvas: TDrawObject); virtual;
    procedure xLeft(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); virtual;
    procedure xRight(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); virtual;

  end;

  TRoofTop = class(TRoof)
  public
    constructor CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
      const cc: TColor;
      const ss: Boolean;
      const IsBasePlate: Boolean;
      const IsTransp: Boolean;
      const LDrawId: string); override;
    {    procedure   Draw(const DrawCanvas: TDrawObject);                      override;}
  protected
    procedure xFront(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); override;
    procedure xBack(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); override;
    procedure xTop(const x, y, z, l, b, h: integer; const st: Boolean; const DrawCanvas: TDrawObject); override;
    {    procedure   xBottom(const x,y,z,l,b,h: integer; const st: Boolean; const DrawCanvas: TDrawObject); override;}
    procedure xLeft(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); override;
    procedure xRight(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); override;
  end;

  TRoofCorner = class(TRoof)
  public
    constructor CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
      const cc: TColor;
      const ss: Boolean;
      const IsBasePlate: Boolean;
      const IsTransp: Boolean;
      const LDrawId: string); override;
    {    procedure   Draw(const DrawCanvas: TDrawObject);                      override;}
    procedure TopStuds(const DrawCanvas: TDrawObject); override;
  protected
    procedure xFront(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); override;
    procedure xBack(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); override;
    procedure xTop(const x, y, z, l, b, h: integer; const st: Boolean; const DrawCanvas: TDrawObject); override;
    {    procedure   xBottom(const x,y,z,l,b,h: integer; const st: Boolean; const DrawCanvas: TDrawObject); override;}
    procedure xLeft(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); override;
    {    procedure   xRight(const x,y,z,l,b,h: integer; const DrawCanvas: TDrawObject);  override;}
  end;

  TRoofTopEnd = class(TRoofTop)
  public
    constructor CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
      const cc: TColor;
      const ss: Boolean;
      const IsBasePlate: Boolean;
      const IsTransp: Boolean;
      const LDrawId: string); override;
    {    procedure   Draw(const DrawCanvas: TDrawObject);                      override;}
    {    procedure   TopStuds(const DrawCanvas: TDrawObject);                  override;}
  protected
    procedure xFront(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); override;
    procedure xBack(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); override;
    procedure xTop(const x, y, z, l, b, h: integer; const st: Boolean; const DrawCanvas: TDrawObject); override;
    {    procedure   xBottom(const x,y,z,l,b,h: integer; const st: Boolean; const DrawCanvas: TDrawObject); override;}
    procedure xLeft(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); override;
    {    procedure   xRight(const x,y,z,l,b,h: integer; const DrawCanvas: TDrawObject);  override;}
  end;

  TRoofTopCorner = class(TRoofCorner)
  public
    constructor CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
      const cc: TColor;
      const ss: Boolean;
      const IsBasePlate: Boolean;
      const IsTransp: Boolean;
      const LDrawId: string); override;
    {    procedure   Draw(const DrawCanvas: TDrawObject);                      override;}
    {    procedure   TopStuds(const DrawCanvas: TDrawObject);                  override;}
  protected
    procedure xFront(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); override;
    procedure xBack(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); override;
    procedure xTop(const x, y, z, l, b, h: integer; const st: Boolean; const DrawCanvas: TDrawObject); override;
    {    procedure   xBottom(const x,y,z,l,b,h: integer; const st: Boolean; const DrawCanvas: TDrawObject); override;}
    procedure xLeft(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); override;
    procedure xRight(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject); override;
  end;

  TRoundBlock = class(TBlock)
    constructor CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
      const cc: TColor;
      const ss: Boolean;
      const IsBasePlate: Boolean;
      const IsTransp: Boolean;
      const LDrawId: string); override;
    procedure Turn; override;
    procedure Draw(const DrawCanvas: TDrawObject); override;
    procedure TopStuds(const DrawCanvas: TDrawObject); override;
  end;

type
  TTrfArr = array[0..3] of TMyFloat;
  TArr1 = array[0..0] of TPoint;
  TArr2 = array[0..1] of TPoint;
  TArr3 = array[0..2] of TPoint;
  TArr4 = array[0..3] of TPoint;
  TArr5 = array[0..4] of TPoint;
  TArr6 = array[0..5] of TPoint;
  TArr7 = array[0..6] of TPoint;
  TArr8 = array[0..7] of TPoint;
  TArr9 = array[0..8] of TPoint;
  TArr100 = array[0..99] of TPoint;
  PArr1 = ^TArr1;
  PArr2 = ^TArr2;
  PArr3 = ^TArr3;
  PArr4 = ^TArr4;
  PArr5 = ^TArr5;
  PArr6 = ^TArr6;
  PArr7 = ^TArr7;
  PArr8 = ^TArr8;
  PArr9 = ^TArr9;
var
  Octant            : integer;
  rx, ry, rz        : TMyFloat;
  tr00, tr01, tr02, tr10, tr11, tr12: TTrfArr;
  xf, yf            : integer;
  xf2, yf2          : integer;
  zx, zy            : integer;
  zx1, zy1          : integer;
  P3D100            : array[0..99] of TPoint3D;
  Pts100            : TArr100;
  Pts9              : TArr9;
  Pts8              : TArr8;
  Pts7              : TArr7;
  Pts6              : TArr6;
  Pts5              : TArr5;
  Pts4              : TArr4;
  Pts3              : TArr3;
  Pts2              : TArr2;
  Pts1              : TArr1;
  Cx, Cy, Cz        : TMyFloat;
  Sx, Sy, Sz        : TMyFloat;
  DetailLevel       : integer;

procedure CalcTransform;
procedure SaveTransform;
procedure RecallTransform;
procedure PrepareRegion(const px, py, pz, pl, pb, ph: integer);
function EdgeColor(const cl: TColor): TColor;

function Point3D(const x, y, z: integer): TPoint3D;

type
  StudSwitchProc = procedure(const x, y: integer; const DrawCanvas: TDrawObject);

var
  ExeLib            : string;
  StudSwitcher      : StudSwitchProc;
  PLX, PLY, PLZ     : ZPLANE;
  PLSX, PLSY        : ZPLANE;
  PLSX1, PLSY1      : ZPLANE;
  lfx, lfy, lfz     : TMyFloat;
  lfsx, lfsy        : TMyFloat;
  lfsx1, lfsy1      : TMyFloat;
  DoShade           : Boolean;
  shx, shy, shz     : TColor;
  shsx, shsy, shtr  : TColor;
  shsx1, shsy1      : TColor;


const
  Light             : TPoint3d = (x: - 7; y: 0; z: 10);
  EdgeShade         : TMyFloat = 0.6;
  Threshold         : TMyFloat = 0.80;

procedure calc_zplane(var zpl: ZPLANE; const p1, p2, p3: TPoint3D);
procedure CalcStandardShades;
procedure Shade(const c: TColor; const f: TMyFloat; var Res: TColor);
procedure ShadeEdge(const c1, c2: TColor; const f: TMyFloat; var Res: TColor);

implementation

uses
  WinProcs, Math, SysUtils, Dialogs,
  Cutting, Main, Lcadpc1, LcadGnrl, CEJUtils;

procedure calc_zplane(var zpl: ZPLANE; const p1, p2, p3: TPoint3D);
begin
  zpl.A := p1.y * (p2.z - p3.z) +
    p2.y * (p3.z - p1.z) +
    p3.y * (p1.z - p2.z);
  zpl.B := p1.z * (p2.x - p3.x) +
    p2.z * (p3.x - p1.x) +
    p3.z * (p1.x - p2.x);
  zpl.C := (-1) * (p1.x * (p2.y - p3.y) +
    p2.x * (p3.y - p1.y) +
    p3.x * (p1.y - p2.y));
  zpl.D := (-1) * p1.x * (p2.y * p3.z - p3.y * p2.z) -
    p2.x * (p3.y * p1.z - p1.y * p3.z) -
    p3.x * (p1.y * p2.z - p2.y * p1.z);
end;

function EdgeColor(const cl: TColor): TColor;
begin
  if (cl = BlackColor) then
    Result := LineColorL
  else if (cl = BlueColor) then
    Result := LineColorB
  else
    Result := LineColorD;
end;

function Point3D(const x, y, z: integer): TPoint3D;
begin
  Result.x := x;
  Result.y := y;
  Result.z := z;
end;

procedure CalcTransform;
var
  i                 : integer;
begin
  Cx := Cos(rx);
  Cy := Cos(ry);
  Sx := Sin(rx);
  Sy := Sin(ry);
  for i := 0 to 3 do begin
    cz := Cos(rz + i * pi / 2);
    sz := Sin(rz + i * pi / 2);
    (**)
    tr00[i] := Cz;
    tr01[i] := Sz;
    tr02[i] := 0;

    tr10[i] := Sz * Cx;
    tr11[i] := -Cz * Cx;
    tr12[i] := Sx;
    (**)
  end;

  Octant := Floor(rz / pi * 2.0) mod 4; {}
  if (rx > pi / 2) then
    Inc(Octant, 4);

  //  MainForm.Edit15.Text := IntToStr(Octant);
end;

var
  sOctant           : integer;
  srx, sry,
    srz, scx        : TMyFloat;
  str00, str01,
    str02, str10,
    str11, str12    : TTrfArr;
  sXBase, sYBase,
    sScale          : TMyFloat;

procedure SaveTransform;
begin
  sXBase := Settings.XBase;
  sYBase := Settings.YBase;
  sScale := Settings.Scale;
  str00 := tr00;
  str01 := tr01;
  str02 := tr02;
  str10 := tr10;
  str11 := tr11;
  str12 := tr12;
  srx := rx;
  sry := ry;
  srz := rz;
  scx := cx;
  sOctant := Octant;
end;

procedure RecallTransform;
begin
  Settings.XBase := sXBase;
  Settings.YBase := sYBase;
  Settings.Scale := sScale;
  tr00 := str00;
  tr01 := str01;
  tr02 := str02;
  tr10 := str10;
  tr11 := str11;
  tr12 := str12;
  rx := srx;
  ry := sry;
  rz := srz;
  cx := scx;
  Octant := sOctant;
end;


constructor TBlock.CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
  const cc: TColor;
  const ss: Boolean;
  const IsBasePlate: Boolean;
  const IsTransp: Boolean;
  const LDrawId: string);
begin
  inherited Create;
  t := 'B';
  px := xx * PFACTORXY;
  py := yy * PFACTORXY;
  pz := zz * PFACTORZ div 3;
  pl := ll * PFACTORXY;
  pb := bb * PFACTORXY;
  ph := hh * PFACTORZ div 3;
  cl := cc;
  st := ss;
  bp := IsBasePlate;
  Transp := IsTransp;
  gpPointer := nil;
  PartId := LDrawId;
  Step := 0;
end;

constructor TRoundBlock.CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
  const cc: TColor;
  const ss: Boolean;
  const IsBasePlate: Boolean;
  const IsTransp: Boolean;
  const LDrawId: string);
begin
  inherited CreateBlock(xx, yy, zz, ll, bb, hh, cc, ss, IsBasePlate, IsTransp, LDrawId);
  t := 'R';
end;

constructor TRoof.CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
  const cc: TColor;
  const ss: Boolean;
  const IsBasePlate: Boolean;
  const IsTransp: Boolean;
  const LDrawId: string);
begin
  inherited CreateBlock(xx, yy, zz, ll, bb, hh, cc, ss, IsBasePlate, IsTransp, LDrawId);
  t := 'T';
  if (ll = 3) and (bb = 2) then
    LDrawPatch := 1;
end;

constructor TRoofTop.CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
  const cc: TColor;
  const ss: Boolean;
  const IsBasePlate: Boolean;
  const IsTransp: Boolean;
  const LDrawId: string);
begin
  inherited CreateBlock(xx, yy, zz, ll, bb, hh, cc, ss, IsBasePlate, IsTransp, LDrawId);
  t := 'N';
  LDrawPatch := 0;
end;

constructor TRoofCorner.CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
  const cc: TColor;
  const ss: Boolean;
  const IsBasePlate: Boolean;
  const IsTransp: Boolean;
  const LDrawId: string);
begin
  inherited CreateBlock(xx, yy, zz, ll, bb, hh, cc, ss, IsBasePlate, IsTransp, LDrawId);
  t := 'C';
end;

constructor TRoofTopCorner.CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
  const cc: TColor;
  const ss: Boolean;
  const IsBasePlate: Boolean;
  const IsTransp: Boolean;
  const LDrawId: string);
begin
  inherited CreateBlock(xx, yy, zz, ll, bb, hh, cc, ss, IsBasePlate, IsTransp, LDrawId);
  t := 'I';
end;

constructor TRoofTopEnd.CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
  const cc: TColor;
  const ss: Boolean;
  const IsBasePlate: Boolean;
  const IsTransp: Boolean;
  const LDrawId: string);
begin
  inherited CreateBlock(xx, yy, zz, ll, bb, hh, cc, ss, IsBasePlate, IsTransp, LDrawId);
  t := 'E';
end;

class function TBlock.CreateCopy(const AnotherBlock: TBlock): TBlock;
begin
  Result := TBlock((AnotherBlock.ClassType).Create);
  with Result do begin
    t := AnotherBlock.t;
    px := AnotherBlock.px;
    py := AnotherBlock.py;
    pz := AnotherBlock.pz;
    pl := AnotherBlock.pl;
    pb := AnotherBlock.pb;
    ph := AnotherBlock.ph;
    cl := AnotherBlock.cl;
    Transp := AnotherBlock.Transp;
    st := AnotherBlock.st;
    bp := AnotherBlock.bp;
    dir := AnotherBlock.Dir;
    gpPointer := AnotherBlock.gpPointer;
    PartId := AnotherBlock.PartId;
    LDrawPatch := AnotherBlock.LDrawPatch;
    Group := 0;
    Step := AnotherBlock.Step;
  end;
end;

{ TBlock }

procedure TBlock.Initialize(const FileName, PartName: string);
begin
end;

function TBlock.ZLevel(const xx, yy, ll, bb: integer): integer;
begin
  if (px < (xx + ll)) and ((px + pl) > xx) and
    (py < (yy + bb)) and ((py + pb) > yy) then
    Result := pz + ph
  else
    Result := 0;
end;

procedure TBlock.Turn;
var
  tmp               : integer;
begin
  Dir := (Dir + 1) and 3;
  tmp := pl;
  pl := pb;
  pb := tmp;
end;

function TBlock.ZeroPoint: TPoint3D;
begin
  Result.x := px;
  Result.y := py;
  Result.z := pz;
end;

procedure PrepareRegion(const px, py, pz, pl, pb, ph: integer);
begin
  case Octant of
    0:
      Transform([Point3D(px, py, pz), Point3D(px, py, pz + ph),
        Point3D(px + pl, py, pz + ph), Point3D(px + pl, py + pb, pz + ph),
          Point3D(px + pl, py + pb, pz), Point3D(px, py + pb, pz)],
          Pts6, 0);
    1:
      Transform([Point3D(px, py, pz), Point3D(px, py + pb, pz),
        Point3D(px, py + pb, pz + ph), Point3D(px + pl, py + pb, pz + ph),
          Point3D(px + pl, py, pz + ph), Point3D(px + pl, py, pz)],
          Pts6, 0);
    2:
      Transform([Point3D(px, py, pz), Point3D(px, py, pz + ph),
        Point3D(px, py + pb, pz + ph), Point3D(px + pl, py + pb, pz + ph),
          Point3D(px + pl, py + pb, pz), Point3D(px + pl, py, pz)],
          Pts6, 0);
    3:
      Transform([Point3D(px, py + pb, pz), Point3D(px + pl, py + pb, pz),
        Point3D(px + pl, py, pz), Point3D(px + pl, py, pz + ph),
          Point3D(px, py, pz + ph), Point3D(px, py + pb, pz + ph)],
          Pts6, 0);
    4:
      Transform([Point3D(px, py, pz), Point3D(px, py, pz + ph),
        Point3D(px, py + pb, pz + ph), Point3D(px + pl, py + pb, pz + ph),
          Point3D(px + pl, py + pb, pz), Point3D(px + pl, py, pz)],
          Pts6, 0);
    5:
      Transform([Point3D(px, py + pb, pz), Point3D(px, py + pb, pz + ph),
        Point3D(px, py, pz + ph), Point3D(px + pl, py, pz + ph),
          Point3D(px + pl, py, pz), Point3D(px + pl, py + pb, pz)],
          Pts6, 0);
    6:
      Transform([Point3D(px, py, pz), Point3D(px, py, pz + ph),
        Point3D(px + pl, py, pz + ph), Point3D(px + pl, py + pb, pz + ph),
          Point3D(px + pl, py + pb, pz), Point3D(px, py + pb, pz)],
          Pts6, 0);
    7:
      Transform([Point3D(px, py + pb, pz + ph), Point3D(px, py + pb, pz),
        Point3D(px, py, pz), Point3D(px + pl, py, pz),
          Point3D(px + pl, py, pz + ph), Point3D(px + pl, py + pb, pz + ph)],
          Pts6, 0);
  end;
end;

function TBlock.HitTest(const xx, yy: integer): Boolean;
var
  Rgn               : HRgn;
begin
  Result := False;
  PrepareRegion(px, py, pz, pl, pb, ph);
  Rgn := CreatePolygonRgn(Pts6, 6, ALTERNATE);
  if (Rgn <> 0) then begin
    try
      Result := PtInRegion(Rgn, xx, yy);
    finally
      DeleteObject(Rgn);
    end
  end
end;

procedure FastStud(const x, y: integer; const DrawCanvas: TDrawObject);
begin
  DrawCanvas.PolyLine([
    Point(x - xf + zx, y + yf2 + zy),
      Point(x - xf + zx, y - yf2 + zy),
      Point(x - xf2 + zx, y - yf + zy),
      Point(x + xf2 + zx, y - yf + zy),
      Point(x + xf + zx, y - yf2 + zy),
      Point(x + xf + zx, y + yf2 + zy),
      Point(x + xf2 + zx, y + yf + zy),
      Point(x - xf2 + zx, y + yf + zy),
      Point(x - xf + zx, y + yf2 + zy), { Now go down  }
    Point(x - xf, y + yf2),
      Point(x - xf2, y + yf),
      Point(x + xf2, y + yf),
      Point(x + xf, y + yf2),           { And up again }
    Point(x + xf + zx, y + yf2 + zy)
      ]);
end;

procedure BetterStud(const x, y: integer; const DrawCanvas: TDrawObject);
begin
  { Bottom 'circle' and two lines upwards }
  DrawCanvas.PolyLine([Point(x - xf, y + yf2),
    Point(x - xf, y - yf2),
      Point(x - xf2, y - yf),
      Point(x + xf2, y - yf),
      Point(x + xf, y - yf2),
      Point(x + xf + zx, y - yf2 + zy),
      Point(x + xf, y - yf2),
      Point(x + xf, y + yf2),
      Point(x + xf2, y + yf),
      Point(x - xf2, y + yf),
      Point(x - xf, y + yf2),
      Point(x - xf + zx, y + yf2 + zy)]);
  { Top surface }
  DrawCanvas.PolyGon([
    Point(x - xf + zx, y + yf2 + zy),
      Point(x - xf + zx, y - yf2 + zy),
      Point(x - xf2 + zx, y - yf + zy),
      Point(x + xf2 + zx, y - yf + zy),
      Point(x + xf + zx, y - yf2 + zy),
      Point(x + xf + zx, y + yf2 + zy),
      Point(x + xf2 + zx, y + yf + zy),
      Point(x - xf2 + zx, y + yf + zy),
      Point(x - xf + zx, y + yf2 + zy)]);
end;

procedure BestStud(const x, y: integer; const DrawCanvas: TDrawObject);
var
  clr               : TColor;
begin
  //  DrawCanvas.Rectangle(x - xf + zx, y + zy, x + xf, y); replace by lines
  DrawCanvas.Ellipse(x - xf, y - yf, x + xf + 1, y + yf);
  clr := DrawCanvas.Canvas.Pen.Color;
  DrawCanvas.Canvas.Pen.Color := DrawCanvas.Canvas.Brush.Color;
  DrawCanvas.Rectangle(x - xf + zx + 1, y + zy, x + xf - 1, y);
  DrawCanvas.Canvas.Pen.Color := clr;
  DrawCanvas.Ellipse(x - xf + zx, y - yf + zy, x + xf + zx + 1, y + yf + zy);
  DrawCanvas.Line(x - xf + zx, y + zy, x - xf + zx, y);
  DrawCanvas.Line(x + xf, y + zy, x + xf, y);
end;

class procedure TBlock.SetStudQuality(const q: integer);
begin
  case q of
    1: StudSwitcher := BetterStud;
    2: StudSwitcher := BestStud;
    else
      StudSwitcher := FastStud;
  end;
end;

class function TBlock.GetStudQuality: integer;
begin
  if @StudSwitcher = @BetterStud then
    Result := 1
  else if @StudSwitcher = @BestStud then
    Result := 2
  else
    Result := 0
end;

procedure TBlock.TopStuds(const DrawCanvas: TDrawObject);
var
  i, j              : integer;
  dl, db            : integer;
  dlInc, dbInc, dbStart: integer;
  x1, y1, z1        : integer;
  p                 : integer;
  sa                : array of TPoint3D;
  sb                : array of TPoint;
  //  l, b              : integer;
begin
  SetLength(sa, pl * pb div PFACTORXY div PFACTORXY + 1);
  SetLength(sb, pl * pb div PFACTORXY div PFACTORXY + 1);
  x1 := px + PFACTORXY div 2;
  y1 := py + PFACTORXY div 2;
  z1 := pz + ph;
  p := 0;
  //  l := pl div PFACTORXY;
  //  b := pb div PFACTORXY;

  case Octant of
    0: begin
        dlInc := -PFACTORXY;
        dl := pl + dlInc;
        dbInc := PFACTORXY;
        dbStart := 0;
      end;
    1: begin
        dlInc := -PFACTORXY;
        dl := pl + dlInc;
        dbInc := -PFACTORXY;
        dbStart := pb + dbInc;
      end;
    2: begin
        dl := 0;
        dlInc := PFACTORXY;
        dbInc := -PFACTORXY;
        dbStart := pb + dbInc;
      end;
    3: begin
        dl := 0;
        dlInc := PFACTORXY;
        dbStart := 0;
        dbInc := PFACTORXY;
      end;
    else begin
        Exit;
      end;
  end;

  for i := 1 to (pl div PFACTORXY) do begin
    db := dbStart;
    for j := 1 to (pb div PFACTORXY) do begin
      (*
            if p > 99 then begin
              Transform(Slice(P3D100, p), Pts100, 0);
              Dec(p);
              while (p >= 0) do begin
                with Pts100[p] do begin
                  StudSwitcher(x, y, DrawCanvas);
                end;
                Dec(p);
              end;
              p := 0;
            end;
      *)
      //      P3D100[p] := Point3D(x1 + dl, y1 + db, z1);
      sa[p] := Point3D(x1 + dl, y1 + db, z1);
      Inc(p);
      Inc(db, dbInc);
    end;

    Inc(dl, dlInc);
  end;

  if (p > 0) then begin
    //    Transform(Slice(P3D100, p), Pts100, 0);
    Transform(sa, sb, 0);
    Dec(p);
    //    while (p >= 0) do begin
    //      with Pts100[p] do begin
    //        StudSwitcher(x, y, DrawCanvas);
    //      end;
    //      Dec(p);
    //    end
    for i := 0 to p do
      StudSwitcher(sb[i].x, sb[i].y, DrawCanvas);
  end
end;

procedure TBlock.BottomStuds(const DrawCanvas: TDrawObject);
var
  i, j              : integer;
  dl, db            : integer;
  x1, y1            : integer;
begin
  dl := PFACTORXY div 2;
  x1 := px + (PFACTORXY div 2);
  y1 := py + (PFACTORXY div 2);
  for i := 1 to Pred(pl div PFACTORXY) do begin
    db := (PFACTORXY div 2);
    for j := 1 to Pred(pb div PFACTORXY) do begin
      Transform([Point3D(x1 + dl, y1 + db, pz)], Pts4, 0);
      with Pts4[0] do begin

        DrawCanvas.PolyLine([Point(x - xf, y + yf2),
          Point(x - xf, y - yf2),
            Point(x - xf2, y - yf),
            Point(x + xf2, y - yf),
            Point(x + xf, y - yf2),
            Point(x + xf, y + yf2),
            Point(x + xf2, y + yf),
            Point(x - xf2, y + yf),
            Point(x - xf, y + yf2)]);
      end;
      Inc(db, PFACTORXY);
    end;
    Inc(dl, PFACTORXY);
  end;
end;

procedure TBlock.Octant0(const DrawCanvas: TDrawObject);
var
  p6                : PArr6;
  p4                : PArr4;
  p3                : PArr3;
  p2                : PArr2;
begin
  if (DoShade) then begin
    P3D100[0] := Point3D(px, py, pz);
    P3D100[1] := Point3D(px, py + pb, pz);
    P3D100[2] := Point3D(px, py + pb, pz + ph);
    P3D100[3] := Point3D(px, py, pz + ph);

    P3D100[4] := Point3D(px + pl, py, pz + ph);
    P3D100[5] := Point3D(px + pl, py + pb, pz + ph);

    P3D100[6] := Point3D(px, py + pb, pz);
    P3D100[7] := Point3D(px, py + pb, pz + ph);
    P3D100[8] := Point3D(px + pl, py + pb, pz + ph);
    P3D100[9] := Point3D(px + pl, py + pb, pz);

    Transform(Slice(P3D100, 10), Pts100, 0);

    p4 := @Pts100[0];
    DrawCanvas.Canvas.Brush.Color := shy;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[2];
    DrawCanvas.Canvas.Brush.Color := shz;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[6];
    DrawCanvas.Canvas.Brush.Color := shx;
    DrawCanvas.Polygon(p4^);
    DrawCanvas.Canvas.Brush.Color := shz;
  end
  else begin
    Transform([Point3D(px, py, pz), Point3D(px, py, pz + ph),
      Point3D(px + pl, py, pz + ph), Point3D(px + pl, py + pb, pz + ph),
        Point3D(px + pl, py + pb, pz), Point3D(px, py + pb, pz),

      Point3D(px, py, pz + ph), Point3D(px, py + pb, pz + ph),
        Point3D(px + pl, py + pb, pz + ph),

      Point3D(px, py + pb, pz + ph), Point3D(px, py + pb, pz)],
        Pts100, 0);
    p6 := @Pts100[0];
    p3 := @Pts100[6];
    p2 := @Pts100[9];
    DrawCanvas.Polygon(p6^);
    DrawCanvas.PolyLine(p3^);
    DrawCanvas.PolyLine(p2^);
  end
end;

procedure TBlock.Octant1(const DrawCanvas: TDrawObject);
var
  p6                : PArr6;
  p4                : PArr4;
  p3                : PArr3;
  p2                : PArr2;
begin
  if (DoShade) then begin
    P3D100[0] := Point3D(px, py, pz);
    P3D100[1] := Point3D(px + pl, py, pz);
    P3D100[2] := Point3D(px + pl, py, pz + ph);
    P3D100[3] := Point3D(px, py, pz + ph);

    P3D100[4] := Point3D(px, py + pb, pz + ph);
    P3D100[5] := Point3D(px + pl, py + pb, pz + ph);

    P3D100[6] := Point3D(px, py, pz);
    P3D100[7] := Point3D(px, py + pb, pz);
    P3D100[8] := Point3D(px, py + pb, pz + ph);
    P3D100[9] := Point3D(px, py, pz + ph);

    Transform(Slice(P3D100, 10), Pts100, 0);

    p4 := @Pts100[0];
    DrawCanvas.Canvas.Brush.Color := shx;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[2];
    DrawCanvas.Canvas.Brush.Color := shz;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[6];
    DrawCanvas.Canvas.Brush.Color := shy;
    DrawCanvas.Polygon(p4^);
    DrawCanvas.Canvas.Brush.Color := shz;
  end
  else begin
    Transform([Point3D(px, py, pz), Point3D(px, py + pb, pz),
      Point3D(px, py + pb, pz + ph), Point3D(px + pl, py + pb, pz + ph),
        Point3D(px + pl, py, pz + ph), Point3D(px + pl, py, pz),

      Point3D(px + pl, py, pz + ph), Point3D(px, py, pz + ph),
        Point3D(px, py + pb, pz + ph),

      Point3D(px, py, pz + ph), Point3D(px, py, pz)],
        Pts100, 0);
    p6 := @Pts100[0];
    p3 := @Pts100[6];
    p2 := @Pts100[9];
    DrawCanvas.Polygon(p6^);
    DrawCanvas.PolyLine(p3^);
    DrawCanvas.PolyLine(p2^);
  end
end;

procedure TBlock.Octant2(const DrawCanvas: TDrawObject);
var
  p6                : PArr6;
  p4                : PArr4;
  p3                : PArr3;
  p2                : PArr2;
begin
  if (DoShade) then begin
    P3D100[0] := Point3D(px + pl, py, pz);
    P3D100[1] := Point3D(px + pl, py + pb, pz);
    P3D100[2] := Point3D(px + pl, py + pb, pz + ph);
    P3D100[3] := Point3D(px + pl, py, pz + ph);

    P3D100[4] := Point3D(px, py, pz + ph);
    P3D100[5] := Point3D(px, py + pb, pz + ph);

    P3D100[6] := Point3D(px, py, pz);
    P3D100[7] := Point3D(px, py, pz + ph);
    P3D100[8] := Point3D(px + pl, py, pz + ph);
    P3D100[9] := Point3D(px + pl, py, pz);

    Transform(Slice(P3D100, 10), Pts100, 0);

    p4 := @Pts100[0];
    DrawCanvas.Canvas.Brush.Color := shy;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[2];
    DrawCanvas.Canvas.Brush.Color := shz;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[6];
    DrawCanvas.Canvas.Brush.Color := shx;
    DrawCanvas.Polygon(p4^);
    DrawCanvas.Canvas.Brush.Color := shz;
  end
  else begin
    Transform([Point3D(px, py, pz), Point3D(px, py, pz + ph),
      Point3D(px, py + pb, pz + ph), Point3D(px + pl, py + pb, pz + ph),
        Point3D(px + pl, py + pb, pz), Point3D(px + pl, py, pz),

      Point3D(px, py, pz + ph), Point3D(px + pl, py, pz + ph),
        Point3D(px + pl, py + pb, pz + ph),

      Point3D(px + pl, py, pz + ph), Point3D(px + pl, py, pz)],
        Pts100, 0);
    p6 := @Pts100[0];
    p3 := @Pts100[6];
    p2 := @Pts100[9];
    DrawCanvas.Polygon(p6^);
    DrawCanvas.PolyLine(p3^);
    DrawCanvas.PolyLine(p2^);
  end
end;

procedure TBlock.Octant3(const DrawCanvas: TDrawObject);
var
  p6                : PArr6;
  p4                : PArr4;
  p3                : PArr3;
  p2                : PArr2;
begin
  if (DoShade) then begin
    P3D100[0] := Point3D(px, py + pb, pz);
    P3D100[1] := Point3D(px + pl, py + pb, pz);
    P3D100[2] := Point3D(px + pl, py + pb, pz + ph);
    P3D100[3] := Point3D(px, py + pb, pz + ph);

    P3D100[4] := Point3D(px, py, pz + ph);
    P3D100[5] := Point3D(px + pl, py, pz + ph);

    P3D100[6] := Point3D(px + pl, py, pz);
    P3D100[7] := Point3D(px + pl, py + pb, pz);
    P3D100[8] := Point3D(px + pl, py + pb, pz + ph);
    P3D100[9] := Point3D(px + pl, py, pz + ph);

    Transform(Slice(P3D100, 10), Pts100, 0);

    p4 := @Pts100[0];
    DrawCanvas.Canvas.Brush.Color := shx;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[2];
    DrawCanvas.Canvas.Brush.Color := shz;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[6];
    DrawCanvas.Canvas.Brush.Color := shy;
    DrawCanvas.Polygon(p4^);
    DrawCanvas.Canvas.Brush.Color := shz;
  end
  else begin
    Transform([Point3D(px, py + pb, pz), Point3D(px + pl, py + pb, pz),
      Point3D(px + pl, py, pz), Point3D(px + pl, py, pz + ph),
        Point3D(px, py, pz + ph), Point3D(px, py + pb, pz + ph),

      Point3D(px, py + pb, pz + ph), Point3D(px + pl, py + pb, pz + ph),
        Point3D(px + pl, py, pz + ph),

      Point3D(px + pl, py + pb, pz + ph), Point3D(px + pl, py + pb, pz)],
        Pts100, 0);
    p6 := @Pts100[0];
    p3 := @Pts100[6];
    p2 := @Pts100[9];
    DrawCanvas.Polygon(p6^);
    DrawCanvas.PolyLine(p3^);
    DrawCanvas.PolyLine(p2^);
  end
end;

procedure TBlock.Octant4(const DrawCanvas: TDrawObject);
var
  p6                : PArr6;
  p4                : PArr4;
  p3                : PArr3;
  p2                : PArr2;
begin
  if (DoShade) then begin
    P3D100[0] := Point3D(px, py, pz + ph);
    P3D100[1] := Point3D(px, py + pb, pz + ph);
    P3D100[2] := Point3D(px, py + pb, pz);
    P3D100[3] := Point3D(px, py, pz);

    P3D100[4] := Point3D(px + pl, py, pz);
    P3D100[5] := Point3D(px + pl, py + pb, pz);

    P3D100[6] := Point3D(px, py + pb, pz);
    P3D100[7] := Point3D(px, py + pb, pz + ph);
    P3D100[8] := Point3D(px + pl, py + pb, pz + ph);
    P3D100[9] := Point3D(px + pl, py + pb, pz);

    Transform(Slice(P3D100, 10), Pts100, 0);

    p4 := @Pts100[0];
    DrawCanvas.Canvas.Brush.Color := shy;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[2];
    DrawCanvas.Canvas.Brush.Color := shz;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[6];
    DrawCanvas.Canvas.Brush.Color := shx;
    DrawCanvas.Polygon(p4^);
    DrawCanvas.Canvas.Brush.Color := shz;
  end
  else begin
    Transform([Point3D(px, py, pz), Point3D(px, py, pz + ph),
      Point3D(px, py + pb, pz + ph), Point3D(px + pl, py + pb, pz + ph),
        Point3D(px + pl, py + pb, pz), Point3D(px + pl, py, pz),

      Point3D(px, py, pz), Point3D(px, py + pb, pz),
        Point3D(px + pl, py + pb, pz),

      Point3D(px, py + pb, pz), Point3D(px, py + pb, pz + ph)],
        Pts100, 0);
    p6 := @Pts100[0];
    p3 := @Pts100[6];
    p2 := @Pts100[9];
    DrawCanvas.Polygon(p6^);
    DrawCanvas.PolyLine(p3^);
    DrawCanvas.PolyLine(p2^);
  end
end;

procedure TBlock.Octant5(const DrawCanvas: TDrawObject);
var
  p6                : PArr6;
  p4                : PArr4;
  p3                : PArr3;
  p2                : PArr2;
begin
  if (DoShade) then begin
    P3D100[0] := Point3D(px, py, pz + ph);
    P3D100[1] := Point3D(px + pl, py, pz + ph);
    P3D100[2] := Point3D(px + pl, py, pz);
    P3D100[3] := Point3D(px, py, pz);

    P3D100[4] := Point3D(px, py + pb, pz);
    P3D100[5] := Point3D(px + pl, py + pb, pz);

    P3D100[6] := Point3D(px, py, pz);
    P3D100[7] := Point3D(px, py + pb, pz);
    P3D100[8] := Point3D(px, py + pb, pz + ph);
    P3D100[9] := Point3D(px, py, pz + ph);

    Transform(Slice(P3D100, 10), Pts100, 0);

    p4 := @Pts100[0];
    DrawCanvas.Canvas.Brush.Color := shx;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[2];
    DrawCanvas.Canvas.Brush.Color := shz;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[6];
    DrawCanvas.Canvas.Brush.Color := shy;
    DrawCanvas.Polygon(p4^);
    DrawCanvas.Canvas.Brush.Color := shz;
  end
  else begin
    Transform([Point3D(px, py + pb, pz), Point3D(px, py + pb, pz + ph),
      Point3D(px, py, pz + ph), Point3D(px + pl, py, pz + ph),
        Point3D(px + pl, py, pz), Point3D(px + pl, py + pb, pz),

      Point3D(px + pl, py, pz), Point3D(px, py, pz),
        Point3D(px, py + pb, pz),

      Point3D(px, py, pz), Point3D(px, py, pz + ph)],
        Pts100, 0);

    p6 := @Pts100[0];
    p3 := @Pts100[6];
    p2 := @Pts100[9];
    DrawCanvas.Polygon(p6^);
    DrawCanvas.PolyLine(p3^);
    DrawCanvas.PolyLine(p2^);
  end
end;

procedure TBlock.Octant6(const DrawCanvas: TDrawObject);
var
  p6                : PArr6;
  p4                : PArr4;
  p3                : PArr3;
  p2                : PArr2;
begin
  if (DoShade) then begin
    P3D100[0] := Point3D(px + pl, py, pz + ph);
    P3D100[1] := Point3D(px + pl, py + pb, pz + ph);
    P3D100[2] := Point3D(px + pl, py + pb, pz);
    P3D100[3] := Point3D(px + pl, py, pz);

    P3D100[4] := Point3D(px, py, pz);
    P3D100[5] := Point3D(px, py + pb, pz);

    P3D100[6] := Point3D(px, py, pz);
    P3D100[7] := Point3D(px, py, pz + ph);
    P3D100[8] := Point3D(px + pl, py, pz + ph);
    P3D100[9] := Point3D(px + pl, py, pz);

    Transform(Slice(P3D100, 10), Pts100, 0);

    p4 := @Pts100[0];
    DrawCanvas.Canvas.Brush.Color := shy;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[2];
    DrawCanvas.Canvas.Brush.Color := shz;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[6];
    DrawCanvas.Canvas.Brush.Color := shx;
    DrawCanvas.Polygon(p4^);
    DrawCanvas.Canvas.Brush.Color := shz;
  end
  else begin
    Transform([Point3D(px, py, pz), Point3D(px, py, pz + ph),
      Point3D(px + pl, py, pz + ph), Point3D(px + pl, py + pb, pz + ph),
        Point3D(px + pl, py + pb, pz), Point3D(px, py + pb, pz),

      Point3D(px, py, pz), Point3D(px + pl, py, pz),
        Point3D(px + pl, py + pb, pz),

      Point3D(px + pl, py, pz), Point3D(px + pl, py, pz + ph)],
        Pts100, 0);

    p6 := @Pts100[0];
    p3 := @Pts100[6];
    p2 := @Pts100[9];
    DrawCanvas.Polygon(p6^);
    DrawCanvas.PolyLine(p3^);
    DrawCanvas.PolyLine(p2^);
  end
end;

procedure TBlock.Octant7(const DrawCanvas: TDrawObject);
var
  p6                : PArr6;
  p4                : PArr4;
  p3                : PArr3;
  p2                : PArr2;
begin
  if (DoShade) then begin
    P3D100[0] := Point3D(px, py + pb, pz + ph);
    P3D100[1] := Point3D(px + pl, py + pb, pz + ph);
    P3D100[2] := Point3D(px + pl, py + pb, pz);
    P3D100[3] := Point3D(px, py + pb, pz);

    P3D100[4] := Point3D(px, py, pz);
    P3D100[5] := Point3D(px + pl, py, pz);

    P3D100[6] := Point3D(px + pl, py, pz);
    P3D100[7] := Point3D(px + pl, py + pb, pz);
    P3D100[8] := Point3D(px + pl, py + pb, pz + ph);
    P3D100[9] := Point3D(px + pl, py, pz + ph);

    Transform(Slice(P3D100, 10), Pts100, 0);

    p4 := @Pts100[0];
    DrawCanvas.Canvas.Brush.Color := shx;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[2];
    DrawCanvas.Canvas.Brush.Color := shz;
    DrawCanvas.Polygon(p4^);
    p4 := @Pts100[6];
    DrawCanvas.Canvas.Brush.Color := shy;
    DrawCanvas.Polygon(p4^);
    DrawCanvas.Canvas.Brush.Color := shz;
  end
  else begin
    Transform([Point3D(px, py + pb, pz + ph), Point3D(px, py + pb, pz),
      Point3D(px, py, pz), Point3D(px + pl, py, pz),
        Point3D(px + pl, py, pz + ph), Point3D(px + pl, py + pb, pz + ph),

      Point3D(px, py + pb, pz), Point3D(px + pl, py + pb, pz),
        Point3D(px + pl, py, pz),

      Point3D(px + pl, py + pb, pz), Point3D(px + pl, py + pb, pz + ph)],
        Pts100, 0);

    p6 := @Pts100[0];
    p3 := @Pts100[6];
    p2 := @Pts100[9];
    DrawCanvas.Polygon(p6^);
    DrawCanvas.PolyLine(p3^);
    DrawCanvas.PolyLine(p2^);
  end
end;


procedure TBlock.Draw(const DrawCanvas: TDrawObject);
var
  c                 : TColor;
begin
  if (pl = 0) then Exit;

  if not Selected then begin
    DrawCanvas.Canvas.Brush.Color := cl;
    DrawCanvas.Canvas.Pen.Color := EdgeColor(cl);
  end
  else begin
    DrawCanvas.Canvas.Brush.Color := cl;
    if cl = RedColor then
      DrawCanvas.Canvas.Pen.Color := clWhite
    else
      DrawCanvas.Canvas.Pen.Color := RedColor;
  end;

  DoShade := MainForm.miShading.Checked;

  if (bp and MainForm.WireBase and (Octant > 3)) or
    (MainForm.WireFrame) then begin
    DrawCanvas.Canvas.Brush.Style := bsClear; { Draw as Wire }
    DrawCanvas.Canvas.Pen.Color := cl;  {}
    DoShade := False;
  end
  else if (cl = clAqua) or (Transp) then begin
    DrawCanvas.Canvas.Brush.Style := bsDiagCross; { Transparent piece }
  end;

  if (DoShade) then begin
    ShadeEdge(DrawCanvas.Canvas.Brush.Color, DrawCanvas.Canvas.Pen.Color, EdgeShade, c);
    if not Selected then
      DrawCanvas.Canvas.Pen.Color := c;
    Shade(DrawCanvas.Canvas.Brush.Color, lfx, shx);
    Shade(DrawCanvas.Canvas.Brush.Color, lfy, shy);
    Shade(DrawCanvas.Canvas.Brush.Color, lfz, shz);
  end;

  case Octant of
    0: Octant0(DrawCanvas);
    1: Octant1(DrawCanvas);
    2: Octant2(DrawCanvas);
    3: Octant3(DrawCanvas);
    4: Octant4(DrawCanvas);
    5: Octant5(DrawCanvas);
    6: Octant6(DrawCanvas);
    7: Octant7(DrawCanvas);
  end;

  if (st and not MainForm.btnStuds.Down) then
    if Octant < 4 then begin
      if (TopCover < pl * pb) then
        TopStuds(DrawCanvas);
    end
    else if (BotCover < pl * pb) then
      BottomStuds(DrawCanvas);

  DrawCanvas.Canvas.Brush.Color := clSilver;
  DrawCanvas.Canvas.Brush.Style := bsSolid;
  DrawCanvas.Canvas.Pen.Color := BlackColor;

end;

function TBlock.AsString: string;
begin
  Result := Format('%s:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%s:%d',
    [t, px, py, pz, pl, pb, ph, cl,
    integer(st), integer(bp), Dir, integer(Transp),
      Group, PartID, Step]);
end;

class function TBlock.CreateFromString(const s: string): TBlock;
var
  a                 : TStringList;
begin
  Result := TBlock.Create;
  a := TStringList.Create;
  try
    if (Split(s, a) <> 15) then
      raise Exception.Create('Illegal string to CreateFromString: ' + s);
    Result.t := a[0][1];
    Result.px := StrToInt(a[1]);
    Result.py := StrToInt(a[2]);
    Result.pz := StrToInt(a[3]);
    Result.pl := StrToInt(a[4]);
    Result.pb := StrToInt(a[5]);
    Result.ph := StrToInt(a[6]);
    Result.cl := StrToInt(a[7]);
    Result.st := a[8] = '1';
    Result.bp := a[9] = '1';
    Result.Dir := StrToInt(a[10]);
    Result.Transp := a[11] = '1';
    Result.Group := StrToInt(a[12]);
    Result.Partid := a[13];
    Result.Step := StrToInt(a[14]);
  finally
    a.Free;
  end;
end;

{ TRoof }

function TRoof.HitTest(const xx, yy: integer): Boolean;
var
  Rgn               : HRgn;
begin
  Result := False;
  PrepareRegion(px, py, pz, pl, pb, ph);
  Rgn := CreatePolygonRgn(Pts6, 6, ALTERNATE);
  if (Rgn <> 0) then begin
    try
      Result := PtInRegion(Rgn, xx, yy);
    finally
      DeleteObject(Rgn);
    end
  end
end;

procedure TRoof.TopStuds(const DrawCanvas: TDrawObject);
var
  i                 : integer;
  dl, db            : integer;
  x1, y1, z1        : integer;
  cnt               : integer;
begin
  dl := 0;
  db := 0;
  cnt := 0;
  case Dir of
    0: begin
        cnt := pl div PFACTORXY;
        dl := 0;
        db := 0;
      end;
    1: begin
        cnt := pb div PFACTORXY;
        dl := 0;
        db := PFACTORXY;
      end;
    2: begin
        cnt := pl div PFACTORXY;
        dl := 0;
        db := pb - PFACTORXY;
      end;
    3: begin
        cnt := pb div PFACTORXY;
        dl := pl - PFACTORXY;
        db := PFACTORXY;
      end;
  end;

  x1 := px + (PFACTORXY div 2);
  y1 := py + (PFACTORXY div 2);
  if (Dir and 1) <> 0 then y1 := -y1;
  z1 := pz + ph;
  for i := 1 to cnt do begin
    Transform([Point3D(x1 + dl, y1 + db, z1)], Pts4, 0);
    with Pts4[0] do begin
      StudSwitcher(x, y, DrawCanvas);
    end;
    if (dir and 1) = 0 then
      Inc(dl, PFACTORXY)
    else
      Inc(db, PFACTORXY);
  end;
end;

procedure TRoof.xFront(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y + b, z + (PFACTORZ div 4)), Point3D(x + l, y + b, z + (PFACTORZ div 4)),
    Point3D(x + l, y + PFACTORXY, z + h), Point3D(x, y + PFACTORXY, z + h)],
      Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsy;
  DrawCanvas.Polygon(Pts4);
  Transform([Point3D(x, y + b, z), Point3D(x + l, y + b, z),
    Point3D(x + l, y + b, z + (PFACTORZ div 4)), Point3D(x, y + b, z + (PFACTORZ div 4))],
      Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shy;
  DrawCanvas.Polygon(Pts4);
end;

procedure TRoof.xLeft(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y, z), Point3D(x, y + b, z),
    Point3D(x, y + b, z + (PFACTORZ div 4)), Point3D(x, y + PFACTORXY, z + h),
      Point3D(x, y, z + h)], Pts5, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shx;
  DrawCanvas.Polygon(Pts5);
end;

procedure TRoof.xTop(const x, y, z, l, b, h: integer; const st: Boolean; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y + b, z + (PFACTORZ div 4)), Point3D(x + l, y + b, z + (PFACTORZ div 4)),
    Point3D(x + l, y + PFACTORXY, z + h), Point3D(x, y + PFACTORXY, z + h)],
      Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsy;
  DrawCanvas.Polygon(Pts4);
  Transform([Point3D(x, y, z + h), Point3D(x, y + PFACTORXY, z + h),
    Point3D(x + l, y + PFACTORXY, z + h), Point3D(x + l, y, z + h)],
      Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shz;
  DrawCanvas.Polygon(Pts4);
  if st then TopStuds(DrawCanvas);
end;

procedure TRoof.xBack(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y, z), Point3D(x + l, y, z),
    Point3D(x + l, y, z + h), Point3D(x, y, z + h)],
      Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shy;
  DrawCanvas.Polygon(Pts4);
end;

procedure TRoof.xRight(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x + l, y, z), Point3D(x + l, y + b, z),
    Point3D(x + l, y + b, z + (PFACTORZ div 4)), Point3D(x + l, y + PFACTORXY, z + h),
      Point3D(x + l, y, z + h)], Pts5, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shx;
  DrawCanvas.Polygon(Pts5);
end;

procedure TRoof.xBottom(const x, y, z, l, b, h: integer; const st: Boolean; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y, z), Point3D(x, y + b, z),
    Point3D(x + l, y + b, z), Point3D(x + l, y, z)],
      Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shz;
  DrawCanvas.Polygon(Pts4);
  if (Dir and 1) = 1 then
    py := -py;
  if st then BottomStuds(DrawCanvas);
end;

procedure TRoof.Draw(const DrawCanvas: TDrawObject);
var
  Studs             : Boolean;
  c                 : TColor;
  xpy               : integer;

  procedure Front;
  begin
    case Dir of
      0: xFront(px, py, pz, pl, pb, ph, DrawCanvas);
      1: xLeft(py - pb, px, pz, pb, pl, ph, DrawCanvas);
      2: xBack(-px - pl, -py - pb, pz, pl, pb, ph, DrawCanvas);
      3: xRight(-py, -px - pl, pz, pb, pl, ph, DrawCanvas);
    end;
  end;

  procedure Left;
  begin
    case Dir of
      0: xLeft(px, py, pz, pl, pb, ph, DrawCanvas);
      1: xBack(py - pb, px, pz, pb, pl, ph, DrawCanvas);
      2: xRight(-px - pl, -py - pb, pz, pl, pb, ph, DrawCanvas);
      3: xFront(-py, -px - pl, pz, pb, pl, ph, DrawCanvas);
    end;
  end;

  procedure Top;
  begin
    case Dir of
      0: xTop(px, py, pz, pl, pb, ph, Studs, DrawCanvas);
      1: xTop(py - pb, px, pz, pb, pl, ph, Studs, DrawCanvas);
      2: xTop(-px - pl, -py - pb, pz, pl, pb, ph, Studs, DrawCanvas);
      3: xTop(-py, -px - pl, pz, pb, pl, ph, Studs, DrawCanvas);
    end;
  end;

  procedure Right;
  begin
    case Dir of
      0: xRight(px, py, pz, pl, pb, ph, DrawCanvas);
      1: xFront(py - pb, px, pz, pb, pl, ph, DrawCanvas);
      2: xLeft(-px - pl, -py - pb, pz, pl, pb, ph, DrawCanvas);
      3: xBack(-py, -px - pl, pz, pb, pl, ph, DrawCanvas);
    end;
  end;

  procedure Back;
  begin
    case Dir of
      0: xBack(px, py, pz, pl, pb, ph, DrawCanvas);
      1: xRight(py - pb, px, pz, pb, pl, ph, DrawCanvas);
      2: xFront(-px - pl, -py - pb, pz, pl, pb, ph, DrawCanvas);
      3: xLeft(-py, -px - pl, pz, pb, pl, ph, DrawCanvas);
    end;
  end;

  procedure Bottom;
  begin
    case Dir of
      0: xBottom(px, py, pz, pl, pb, ph, Studs, DrawCanvas);
      1: xBottom(py - pb, px, pz, pb, pl, ph, Studs, DrawCanvas);
      2: xBottom(-px - pl, -py - pb, pz, pl, pb, ph, Studs, DrawCanvas);
      3: xBottom(-py, -px - pl, pz, pb, pl, ph, Studs, DrawCanvas);
    end;
  end;

begin
  xpy := py;

  Studs := st and not MainForm.btnStuds.Down;
  DoShade := MainForm.miShading.Checked;

  DrawCanvas.Canvas.Brush.Color := cl;
  if not Selected then begin
    DrawCanvas.Canvas.Pen.Color := EdgeColor(cl);
  end
  else begin
    if cl = RedColor then
      DrawCanvas.Canvas.Pen.Color := clWhite
    else
      DrawCanvas.Canvas.Pen.Color := RedColor;
  end;

  if (bp and MainForm.WireBase and (Octant > 3)) or
    (MainForm.WireFrame) then begin
    DrawCanvas.Canvas.Brush.Style := bsClear; { Draw as Wire }
    DrawCanvas.Canvas.Pen.Color := cl;  {}
    DoShade := False;
  end
  else if (cl = clAqua) or (Transp) then begin
    DrawCanvas.Canvas.Brush.Style := bsDiagCross; { Transparent piece }
  end;

  shx := DrawCanvas.Canvas.Brush.Color;
  shy := shx;
  shz := shx;
  shsx := shx;
  shsy := shx;

  if (Dir and 1) = 0 then begin
    py := xpy;
    if (DoShade) then begin
      ShadeEdge(DrawCanvas.Canvas.Brush.Color, DrawCanvas.Canvas.Pen.Color, EdgeShade, c);
      if not Selected then
        DrawCanvas.Canvas.Pen.Color := c;
      Shade(DrawCanvas.Canvas.Brush.Color, lfy, shx);
      Shade(DrawCanvas.Canvas.Brush.Color, lfx, shy);
      Shade(DrawCanvas.Canvas.Brush.Color, lfz, shz);
      if (Dir and 2) = 0 then begin     { 0 }
        Shade(DrawCanvas.Canvas.Brush.Color, lfsx1, shsy);
        Shade(DrawCanvas.Canvas.Brush.Color, lfsy, shsx);
        Shade(DrawCanvas.Canvas.Brush.Color, lfsx, shsy1);
        Shade(DrawCanvas.Canvas.Brush.Color, lfsy1, shsx1);
      end
      else begin                        { 2 }
        Shade(DrawCanvas.Canvas.Brush.Color, lfsx, shsy);
        Shade(DrawCanvas.Canvas.Brush.Color, lfsy1, shsx);
        Shade(DrawCanvas.Canvas.Brush.Color, lfsx1, shsy1);
        Shade(DrawCanvas.Canvas.Brush.Color, lfsy, shsx1);
      end
    end
  end
  else begin
    py := -xpy;
    if (DoShade) then begin
      ShadeEdge(DrawCanvas.Canvas.Brush.Color, DrawCanvas.Canvas.Pen.Color, EdgeShade, c);
      if not Selected then
        DrawCanvas.Canvas.Pen.Color := c;
      Shade(DrawCanvas.Canvas.Brush.Color, lfx, shx);
      Shade(DrawCanvas.Canvas.Brush.Color, lfy, shy);
      Shade(DrawCanvas.Canvas.Brush.Color, lfz, shz);
      if (dir and 2) = 0 then begin     { 1 }
        Shade(DrawCanvas.Canvas.Brush.Color, lfsx1, shsx);
        Shade(DrawCanvas.Canvas.Brush.Color, lfsy1, shsy);
        Shade(DrawCanvas.Canvas.Brush.Color, lfsx, shsx1);
        Shade(DrawCanvas.Canvas.Brush.Color, lfsy, shsy1);
      end
      else begin                        { 3 }
        Shade(DrawCanvas.Canvas.Brush.Color, lfsx, shsx);
        Shade(DrawCanvas.Canvas.Brush.Color, lfsy, shsy);
        Shade(DrawCanvas.Canvas.Brush.Color, lfsx1, shsx1);
        Shade(DrawCanvas.Canvas.Brush.Color, lfsy1, shsy1);
      end
    end
  end;

  if (rx = 0) then
    Top
  else
    case Octant of
      0: begin
          Top;
          Left;
          Front;
        end;
      1: begin
          Top;
          Left;
          Back;
        end;
      2: begin
          Top;
          Right;
          Back;
        end;
      3: begin
          Top;
          Right;
          Front;
        end;
      4: begin
          Front;
          Left;
          Bottom;
        end;
      5: begin
          Back;
          Left;
          Bottom;
        end;
      6: begin
          Back;
          Right;
          Bottom;
        end;
      7: begin
          Front;
          Right;
          Bottom;
        end;
    end;

  py := xpy;

  DrawCanvas.Canvas.Brush.Color := clSilver;
  DrawCanvas.Canvas.Brush.Style := bsSolid;
  DrawCanvas.Canvas.Pen.Color := BlackColor;
end;

{ TRoundBlock }

procedure TRoundBlock.Draw(const DrawCanvas: TDrawObject);
var
  hx, hy            : integer;
  //  x1f, y1f:   integer;
  x2f, y2f          : integer;
  Studs             : Boolean;
  clr               : TColor;
begin
  Studs := st and not MainForm.btnStuds.Down;
  DoShade := MainForm.miShading.Checked;

  if not Selected then begin
    DrawCanvas.Canvas.Brush.Color := cl;
    DrawCanvas.Canvas.Pen.Color := EdgeColor(cl);
  end
  else begin
    DrawCanvas.Canvas.Brush.Color := SelColor;
    DrawCanvas.Canvas.Brush.Color := cl;
    if cl = RedColor then
      DrawCanvas.Canvas.Pen.Color := clWhite
    else
      DrawCanvas.Canvas.Pen.Color := RedColor;
  end;

  if (bp and MainForm.WireBase and (Octant > 3)) or
    (MainForm.WireFrame) then begin
    DrawCanvas.Canvas.Brush.Style := bsClear; { Draw as Wire }
    DrawCanvas.Canvas.Pen.Color := cl;  {}
    DoShade := False;
  end
  else if (cl = clAqua) or (Transp) then begin
    DrawCanvas.Canvas.Brush.Style := bsDiagCross; { Transparent piece }
  end;

  if (DoShade) then begin
    ShadeEdge(DrawCanvas.Canvas.Brush.Color, DrawCanvas.Canvas.Pen.Color, EdgeShade, clr);
    if not Selected then
      DrawCanvas.Canvas.Pen.Color := clr;
  end;

  hx := (ph * zx1 div PFACTORZ);
  hy := (ph * zy1 div PFACTORZ);

  x2f := Round(pl / 2.0 * Settings.Scale);
  y2f := Round(pb / 2.0 * Abs(cos(rx)) * Settings.Scale);

  Transform([Point3D(px + (pl div 2), py + (pb div 2), pz)], Pts4, 0);

  with Pts4[0] do begin
    if (Octant < 4) then begin
      (* Better routine, but a bit slow *)
      DrawCanvas.Rectangle(x - x2f + hx, y + hy, x + x2f + 1, y);
      DrawCanvas.Ellipse(x - x2f, y - y2f, x + x2f + 1, y + y2f + 1);
      clr := DrawCanvas.Canvas.Pen.Color;
      DrawCanvas.Canvas.Pen.Color := DrawCanvas.Canvas.Brush.Color;
      DrawCanvas.Rectangle(x - x2f + hx + 1, y + hy, x + x2f, y);
      DrawCanvas.Canvas.Pen.Color := clr;
      DrawCanvas.Ellipse(x - x2f + hx, y - y2f + hy, x + x2f + hx + 1, y + y2f + hy + 1);
      (* Keep the old one, just in case...
            DrawCanvas.PolyGon([Point(x+x2f+hx,y+hy),
                                Point(x+x2f,y),
                                Point(x+x1f,y+y1f),
                                Point(x,y+y2f),
                                Point(x-x1f, y+y1f),
                                Point(x-x2f, y),
                                Point(x-x2f+hx,y+hy),
                                Point(x-x1f+hx,y+y1f+hy),
                                Point(x+hx, y+y2f+hy),
                                Point(x+x1f+hx,y+y1f+hy),
                                Point(x+x2f,y+hy)]);
            DrawCanvas.PolyGon([Point(x-x2f+hx,y+hy),
                                Point(x-x1f+hx,y+y1f+hy),
                                Point(x+hx, y+y2f+hy),
                                Point(x+x1f+hx,y+y1f+hy),
                                Point(x+x2f+hx,y+hy),
                                Point(x+x2f+hx,y+hy),
                                Point(x+x1f+hx,y-y1f+hy),
                                Point(x+hx, y-y2f+hy),
                                Point(x-x1f+hx,y-y1f+hy),
                                Point(x-x2f+hx,y+hy)]);
      *)
      if Studs then TopStuds(DrawCanvas);
    end
    else begin
      DrawCanvas.Rectangle(x - x2f + hx, y + hy, x + x2f + 1, y);
      DrawCanvas.Ellipse(x - x2f + hx, y - y2f + hy, x + x2f + hx + 1, y + y2f + hy + 1);
      clr := DrawCanvas.Canvas.Pen.Color;
      DrawCanvas.Canvas.Pen.Color := DrawCanvas.Canvas.Brush.Color;
      DrawCanvas.Rectangle(x - x2f + hx + 1, y + hy, x + x2f, y);
      DrawCanvas.Canvas.Pen.Color := clr;
      DrawCanvas.Ellipse(x - x2f, y - y2f, x + x2f + 1, y + y2f + 1);
      (*
            DrawCanvas.PolyGon([Point(x+x2f+hx,y+hy),
                                Point(x+x2f,y),
                                Point(x+x1f,y+y1f),
                                Point(x,y+y2f),
                                Point(x-x1f, y+y1f),
                                Point(x-x2f, y),
                                Point(x-x2f+hx,y+hy),
                                Point(x-x1f+hx,y-y1f+hy),
                                Point(x+hx, y-y2f+hy),
                                Point(x+x1f+hx,y-y1f+hy),
                                Point(x+x2f,y+hy)]);
            DrawCanvas.PolyGon([Point(x-x2f,y),
                                Point(x-x1f,y-y1f),
                                Point(x, y-y2f),
                                Point(x+x1f,y-y1f),
                                Point(x+x2f,y),
                                Point(x+x2f,y),
                                Point(x+x1f,y+y1f),
                                Point(x, y+y2f),
                                Point(x-x1f,y+y1f),
                                Point(x-x2f,y)]);
      *)
      DrawCanvas.PolyLine([Point(x - xf, y + yf2),
        Point(x - xf, y - yf2),
          Point(x - xf2, y - yf),
          Point(x + xf2, y - yf),
          Point(x + xf, y - yf2),
          Point(x + xf, y + yf2),
          Point(x + xf2, y + yf),
          Point(x - xf2, y + yf),
          Point(x - xf, y + yf2)]);

      (*      if Studs then BottomStuds(DrawCanvas); no effect on 1x1 *)
    end;
  end;
  DrawCanvas.Canvas.Brush.Color := clSilver;
  DrawCanvas.Canvas.Brush.Style := bsSolid;
  DrawCanvas.Canvas.Pen.Color := BlackColor;
end;

procedure TRoundBlock.TopStuds(const DrawCanvas: TDrawObject);
var
  i, j              : integer;
  dl, db            : integer;
  x1, y1, z1        : integer;
begin
  dl := 0;
  x1 := px + PFACTORXY div 2;
  y1 := py + PFACTORXY div 2;
  z1 := pz + ph;
  for i := 1 to (pl div PFACTORXY) do begin
    db := 0;
    for j := 1 to (pb div PFACTORXY) do begin
      if (pl < 3 * PFACTORXY) or
        ((pl = 3 * PFACTORXY) and ((i = 2) or (j = 2))) or
        ((pl = 4 * PFACTORXY) and ((i in [2, 3]) or (j in [2, 3]))) or
        ((pl = 5 * PFACTORXY) and ((i in [2, 3, 4]) or (j in [2, 3, 4]))) or
        ((pl = 6 * PFACTORXY) and ((i in [3, 4]) or (j in [3, 4]) or ((i in [2, 5]) and (j in [2, 5])))) or
        ((pl = 7 * PFACTORXY) and ((i in [3, 4, 5]) or (j in [3, 4, 5]) or ((i in [2, 6]) and (j in [2, 6])))) or
        ((pl = 8 * PFACTORXY) and ((i in [4, 5]) or (j in [4, 5]) or ((i in [2, 3, 6, 7]) and (j in [2, 3, 6, 7])))) then begin
        Transform([Point3D(x1 + dl, y1 + db, z1)], Pts4, 0);
        with Pts4[0] do begin
          StudSwitcher(x, y, DrawCanvas);
        end;
      end;
      Inc(db, PFACTORXY);
    end;
    Inc(dl, PFACTORXY);
  end;
end;

procedure TRoundBlock.Turn;
begin
  { no action }
end;

{ TRoofTop }

procedure TRoofTop.xFront(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y + b, z + (PFACTORZ div 4)),
    Point3D(x + l, y + b, z + (PFACTORZ div 4)),
      Point3D(x + l, y + b div 2, z + h),
      Point3D(x, y + b div 2, z + h)],
      Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsy;
  DrawCanvas.Polygon(Pts4);
  Transform([Point3D(x, y + b, z),
    Point3D(x + l, y + b, z),
      Point3D(x + l, y + b, z + (PFACTORZ div 4)),
      Point3D(x, y + b, z + (PFACTORZ div 4))],
      Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shy;
  DrawCanvas.Polygon(Pts4);
end;

procedure TRoofTop.xLeft(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y, z),
    Point3D(x, y + b, z),
      Point3D(x, y + b, z + (PFACTORZ div 4)),
      Point3D(x, y + b div 2, z + h),
      Point3D(x, y, z + (PFACTORZ div 4))], Pts5, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shx;
  DrawCanvas.Polygon(Pts5);
end;

procedure TRoofTop.xTop(const x, y, z, l, b, h: integer; const st: Boolean; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y, z + (PFACTORZ div 4)),
    Point3D(x + l, y, z + (PFACTORZ div 4)),
      Point3D(x + l, y + b div 2, z + h),
      Point3D(x, y + b div 2, z + h)], Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsy1;
  DrawCanvas.Polygon(Pts4);
  Transform([Point3D(x, y + b, z + (PFACTORZ div 4)),
    Point3D(x + l, y + b, z + (PFACTORZ div 4)),
      Point3D(x + l, y + b div 2, z + h),
      Point3D(x, y + b div 2, z + h)], Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsy;
  DrawCanvas.Polygon(Pts4);
end;

procedure TRoofTop.xBack(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y, z + (PFACTORZ div 4)),
    Point3D(x + l, y, z + (PFACTORZ div 4)),
      Point3D(x + l, y + b div 2, z + h),
      Point3D(x, y + b div 2, z + h)], Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsy1;
  DrawCanvas.Polygon(Pts4);
  Transform([Point3D(x, y, z),
    Point3D(x + l, y, z),
      Point3D(x + l, y, z + (PFACTORZ div 4)),
      Point3D(x, y, z + (PFACTORZ div 4))],
      Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shy;
  DrawCanvas.Polygon(Pts4);
end;

procedure TRoofTop.xRight(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x + l, y, z),
    Point3D(x + l, y + b, z),
      Point3D(x + l, y + b, z + (PFACTORZ div 4)),
      Point3D(x + l, y + b div 2, z + h),
      Point3D(x + l, y, z + (PFACTORZ div 4))], Pts5, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shx;
  DrawCanvas.Polygon(Pts5);
end;

{ TRoofCorner }

procedure TRoofCorner.TopStuds(const DrawCanvas: TDrawObject);
var
  i                 : integer;
  dl, db            : integer;
  x1, y1, z1        : integer;
  cnt               : integer;
begin
  dl := 0;
  db := 0;
  cnt := 0;
  case Dir of
    0: begin
        cnt := 1;
        dl := pl - PFACTORXY;
        db := 0;
      end;
    1: begin
        cnt := 1;
        dl := 0;
        db := PFACTORXY;
      end;
    2: begin
        cnt := 1;
        dl := 0;
        db := pb - PFACTORXY;
      end;
    3: begin
        cnt := 1;
        dl := pl - PFACTORXY;
        db := pb;
      end;
  end;

  x1 := px + (PFACTORXY div 2);
  y1 := py + (PFACTORXY div 2);
  if (Dir and 1) <> 0 then y1 := -y1;
  z1 := pz + ph;
  for i := 1 to cnt do begin
    Transform([Point3D(x1 + dl, y1 + db, z1)], Pts4, 0);
    with Pts4[0] do begin
      StudSwitcher(x, y, DrawCanvas);
    end;
    if (dir and 1) = 0 then
      Inc(dl, PFACTORXY)
    else
      Inc(db, PFACTORXY);
  end;
end;

procedure TRoofCorner.xFront(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y + b, z + (PFACTORZ div 4)), Point3D(x + l, y + b, z + (PFACTORZ div 4)),
    Point3D(x + l, y + PFACTORXY, z + h), Point3D(x + l - PFACTORXY, y + PFACTORXY, z + h)], Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsy;
  DrawCanvas.Polygon(Pts4);
  Transform([Point3D(x, y + b, z), Point3D(x + l, y + b, z),
    Point3D(x + l, y + b, z + (PFACTORZ div 4)), Point3D(x, y + b, z + (PFACTORZ div 4))], Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shy;
  DrawCanvas.Polygon(Pts4);
end;

procedure TRoofCorner.xLeft(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y, z + (PFACTORZ div 4)), Point3D(x, y + b, z + (PFACTORZ div 4)),
    Point3D(x + l - PFACTORXY, y + PFACTORXY, z + h), Point3D(x + l - PFACTORXY, y, z + h)], Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsx;
  DrawCanvas.Polygon(Pts4);
  Transform([Point3D(x, y, z), Point3D(x, y + b, z),
    Point3D(x, y + b, z + (PFACTORZ div 4)), Point3D(x, y, z + (PFACTORZ div 4))], Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shx;
  DrawCanvas.Polygon(Pts4);
end;

procedure TRoofCorner.xTop(const x, y, z, l, b, h: integer; const st: Boolean; const DrawCanvas: TDrawObject);
begin
  if ((MainForm.RotZ.Position + Dir * 90) < 220) then begin
    Transform([Point3D(x, y + b, z + (PFACTORZ div 4)), Point3D(x + l, y + b, z + (PFACTORZ div 4)),
      Point3D(x + l, y + PFACTORXY, z + h), Point3D(x + l - PFACTORXY, y + PFACTORXY, z + h)], Pts4, Dir);
    if DoShade then
      DrawCanvas.Canvas.Brush.Color := shsy;
    DrawCanvas.Polygon(Pts4);
    Transform([Point3D(x, y, z + (PFACTORZ div 4)), Point3D(x, y + b, z + (PFACTORZ div 4)),
      Point3D(x + l - PFACTORXY, y + PFACTORXY, z + h), Point3D(x + l - PFACTORXY, y, z + h)], Pts4, Dir);
    if DoShade then
      DrawCanvas.Canvas.Brush.Color := shsx;
    DrawCanvas.Polygon(Pts4);
  end
  else begin
    Transform([Point3D(x, y, z + (PFACTORZ div 4)), Point3D(x, y + b, z + (PFACTORZ div 4)),
      Point3D(x + l - PFACTORXY, y + PFACTORXY, z + h), Point3D(x + l - PFACTORXY, y, z + h)], Pts4, Dir);
    if DoShade then
      DrawCanvas.Canvas.Brush.Color := shsx;
    DrawCanvas.Polygon(Pts4);
    Transform([Point3D(x, y + b, z + (PFACTORZ div 4)), Point3D(x + l, y + b, z + (PFACTORZ div 4)),
      Point3D(x + l, y + PFACTORXY, z + h), Point3D(x + l - PFACTORXY, y + PFACTORXY, z + h)], Pts4, Dir);
    if DoShade then
      DrawCanvas.Canvas.Brush.Color := shsy;
    DrawCanvas.Polygon(Pts4);
  end;
  Transform([Point3D(x + l - PFACTORXY, y, z + h), Point3D(x + l - PFACTORXY, y + PFACTORXY, z + h),
    Point3D(x + l, y + PFACTORXY, z + h), Point3D(x + l, y, z + h)], Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shz;
  DrawCanvas.Polygon(Pts4);
  if st then TopStuds(DrawCanvas);
end;

procedure TRoofCorner.xBack(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y, z), Point3D(x + l, y, z),
    Point3D(x + l, y, z + h), Point3D(x + l - PFACTORXY, y, z + h),
      Point3D(x, y, z + (PFACTORZ div 4))], Pts5, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shy;
  DrawCanvas.Polygon(Pts5);
end;

{ TRoofTopCorner }

procedure TRoofTopCorner.xFront(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y + b, z + (PFACTORZ div 4)),
    Point3D(x + l, y + b, z + (PFACTORZ div 4)),
      Point3D(x + l, y + b div 2, z + h),
      Point3D(x + l div 2, y + b div 2, z + h)], Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsy;
  DrawCanvas.Polygon(Pts4);
  Transform([Point3D(x, y + b, z),
    Point3D(x + l, y + b, z),
      Point3D(x + l, y + b, z + (PFACTORZ div 4)),
      Point3D(x, y + b, z + (PFACTORZ div 4))], Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shy;
  DrawCanvas.Polygon(Pts4);
end;

procedure TRoofTopCorner.xLeft(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y, z + (PFACTORZ div 4)),
    Point3D(x, y + b, z + (PFACTORZ div 4)),
      Point3D(x + l div 2, y + b div 2, z + h),
      Point3D(x + l div 2, y, z + h)], Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsx;
  DrawCanvas.Polygon(Pts4);
  Transform([Point3D(x, y, z),
    Point3D(x, y + b, z),
      Point3D(x, y + b, z + (PFACTORZ div 4)),
      Point3D(x, y, z + (PFACTORZ div 4))], Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shx;
  DrawCanvas.Polygon(Pts4);
end;

procedure TRoofTopCorner.xTop(const x, y, z, l, b, h: integer; const st: Boolean; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x + l, y, z + (PFACTORZ div 4)),
    Point3D(x + l div 2, y + b div 2, z + h),
      Point3D(x + l, y + b div 2, z + h)],
      Pts3, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsy1;
  DrawCanvas.Polygon(Pts3);
  Transform([Point3D(x + l, y, z + (PFACTORZ div 4)),
    Point3D(x + l div 2, y + b div 2, z + h),
      Point3D(x + l div 2, y, z + h)],
      Pts3, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsx1;
  DrawCanvas.Polygon(Pts3);
  Transform([Point3D(x, y + b, z + (PFACTORZ div 4)),
    Point3D(x + l, y + b, z + (PFACTORZ div 4)),
      Point3D(x + l, y + b div 2, z + h),
      Point3D(x + l div 2, y + b div 2, z + h)], Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsy;
  DrawCanvas.Polygon(Pts4);
  Transform([Point3D(x, y, z + (PFACTORZ div 4)),
    Point3D(x, y + b, z + (PFACTORZ div 4)),
      Point3D(x + l div 2, y + b div 2, z + h),
      Point3D(x + l div 2, y, z + h)], Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsx;
  DrawCanvas.Polygon(Pts4);
end;

procedure TRoofTopCorner.xBack(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x + l, y, z + (PFACTORZ div 4)),
    Point3D(x + l div 2, y + b div 2, z + h),
      Point3D(x + l, y + b div 2, z + h)],
      Pts3, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsy1;
  DrawCanvas.Polygon(Pts3);
  Transform([Point3D(x, y, z),
    Point3D(x + l, y, z),
      Point3D(x + l, y, z + (PFACTORZ div 4)),
      Point3D(x + l div 2, y, z + h),
      Point3D(x, y, z + (PFACTORZ div 4))], Pts5, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsx1;
  DrawCanvas.Polygon(Pts5);
end;

procedure TRoofTopCorner.xRight(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x + l, y, z + (PFACTORZ div 4)),
    Point3D(x + l div 2, y + b div 2, z + h),
      Point3D(x + l div 2, y, z + h)],
      Pts3, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsx1;
  DrawCanvas.Polygon(Pts3);
  Transform([Point3D(x + l, y, z),
    Point3D(x + l, y + b, z),
      Point3D(x + l, y + b, z + (PFACTORZ div 4)),
      Point3D(x + l, y + b div 2, z + h),
      Point3D(x + l, y, z + (PFACTORZ div 4))], Pts5, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsy;
  DrawCanvas.Polygon(Pts5);
end;

{ TRoofTopEnd }

procedure TRoofTopEnd.xLeft(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y, z + (PFACTORZ div 4)), Point3D(x, y + b, z + (PFACTORZ div 4)),
    Point3D(x + l, y + b div 2, z + h)],
      Pts3, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsx;
  DrawCanvas.Polygon(Pts3);
  Transform([Point3D(x, y, z), Point3D(x, y + b, z),
    Point3D(x, y + b, z + (PFACTORZ div 4)), Point3D(x, y, z + (PFACTORZ div 4))],
      Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shx;
  DrawCanvas.Polygon(Pts4);
end;

procedure TRoofTopEnd.xTop(const x, y, z, l, b, h: integer; const st: Boolean; const DrawCanvas: TDrawObject);
begin
  if ((MainForm.RotZ.Position + Dir * 90) < 220) or
    (((MainForm.RotZ.Position + Dir * 90) > 270) and
    ((MainForm.RotZ.Position + Dir * 90) < 310)) then begin
    Transform([Point3D(x, y + b, z + (PFACTORZ div 4)),
      Point3D(x + l, y + b, z + (PFACTORZ div 4)),
        Point3D(x + l, y + b div 2, z + h)],
        Pts3, Dir);
    if DoShade then
      DrawCanvas.Canvas.Brush.Color := shsy;
    DrawCanvas.Polygon(Pts3);
    Transform([Point3D(x, y, z + (PFACTORZ div 4)),
      Point3D(x, y + b, z + (PFACTORZ div 4)),
        Point3D(x + l, y + b div 2, z + h)],
        Pts3, Dir);
    if DoShade then
      DrawCanvas.Canvas.Brush.Color := shsx;
    DrawCanvas.Polygon(Pts3);
    Transform([Point3D(x, y, z + (PFACTORZ div 4)),
      Point3D(x + l, y, z + (PFACTORZ div 4)),
        Point3D(x + l, y + b div 2, z + h)],
        Pts3, Dir);
    if DoShade then
      DrawCanvas.Canvas.Brush.Color := shsy1;
    DrawCanvas.Polygon(Pts3);
  end
  else begin
    Transform([Point3D(x, y, z + (PFACTORZ div 4)),
      Point3D(x + l, y, z + (PFACTORZ div 4)),
        Point3D(x + l, y + b div 2, z + h)],
        Pts3, Dir);
    if DoShade then
      DrawCanvas.Canvas.Brush.Color := shsy1;
    DrawCanvas.Polygon(Pts3);
    Transform([Point3D(x, y, z + (PFACTORZ div 4)),
      Point3D(x, y + b, z + (PFACTORZ div 4)),
        Point3D(x + l, y + b div 2, z + h)],
        Pts3, Dir);
    if DoShade then
      DrawCanvas.Canvas.Brush.Color := shsx;
    DrawCanvas.Polygon(Pts3);
    Transform([Point3D(x, y + b, z + (PFACTORZ div 4)),
      Point3D(x + l, y + b, z + (PFACTORZ div 4)),
        Point3D(x + l, y + b div 2, z + h)],
        Pts3, Dir);
    if DoShade then
      DrawCanvas.Canvas.Brush.Color := shsy;
    DrawCanvas.Polygon(Pts3);
  end;
end;

procedure TRoofTopEnd.xBack(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y, z + (PFACTORZ div 4)),
    Point3D(x + l, y, z + (PFACTORZ div 4)),
      Point3D(x + l, y + b div 2, z + h)],
      Pts3, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsy1;
  DrawCanvas.Polygon(Pts3);
  Transform([Point3D(x, y, z),
    Point3D(x + l, y, z),
      Point3D(x + l, y, z + (PFACTORZ div 4)),
      Point3D(x, y, z + (PFACTORZ div 4))],
      Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shy;
  DrawCanvas.Polygon(Pts4);
end;

procedure TRoofTopEnd.xFront(const x, y, z, l, b, h: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(x, y + b, z + (PFACTORZ div 4)),
    Point3D(x + l, y + b, z + (PFACTORZ div 4)),
      Point3D(x + l, y + b div 2, z + h)],
      Pts3, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shsy;
  DrawCanvas.Polygon(Pts3);
  Transform([Point3D(x, y + b, z),
    Point3D(x + l, y + b, z),
      Point3D(x + l, y + b, z + (PFACTORZ div 4)),
      Point3D(x, y + b, z + (PFACTORZ div 4))], Pts4, Dir);
  if DoShade then
    DrawCanvas.Canvas.Brush.Color := shy;
  DrawCanvas.Polygon(Pts4);
end;

{ TBlockList }

procedure TBlockList.Clear;
begin
  inherited;
  fUndoList.Clear;
  fRedoList.Clear;
end;

procedure TBlockList.ClearKeepUndo;
begin
  inherited Clear;
end;

constructor TBlockList.Create;
begin
  fUndoList := TStringList.Create;
  fRedoList := TStringList.Create;
end;

destructor TBlockList.Destroy;
begin
  inherited;
  fUndoList.Free;
  fRedoList.Free;
end;

procedure TBlockList.Redo;
var
  Cmd               : char;
  s                 : string;
  i                 : integer;
begin
  if fRedoList.Count > 0 then begin
    Cmd := '[';
    while Cmd = '[' do begin
      s := fRedoList[0];
      Cmd := s[1];
      fRedoList.Delete(0);
      { do the job }
      case Cmd of
        '+', ']': begin
            // Redo an add == add a block
            Add(TGeneralBlock.CreateFromString(Copy(s, 3, MaxInt)));
          end;
        '-', '[': begin
            // Redo a delete == remove a block
            i := LocateAsString(Copy(s, 3, MaxInt));
            if i >= 0 then begin
              TBlock(Items[i]).Free;
              Delete(i);
            end
          end;
        else begin
            raise Exception.Create('Illegal Redo Command: ' + s);
          end
      end;
      fUndoList.Insert(0, s);
      MainForm.btnUndo.Enabled := True;
    end;
  end;
end;

procedure TBlockList.Undo;
var
  Cmd               : char;
  s                 : string;
  i                 : integer;
begin
  if fUndoList.Count > 0 then begin
    Cmd := ']';
    while Cmd = ']' do begin
      s := fUndoList[0];
      Cmd := s[1];
      fUndoList.Delete(0);
      { do the job }
      case Cmd of
        '+', ']': begin
            // Undo an add == remove a block
            i := LocateAsString(Copy(s, 3, MaxInt));
            if i >= 0 then begin
              TBlock(Items[i]).Free;
              Delete(i);
            end
          end;
        '-', '[': begin
            // Undo a delete == add a block
            Add(TGeneralBlock.CreateFromString(Copy(s, 3, MaxInt)));
          end;
        else begin
            raise Exception.Create('Illegal Undo Command: ' + s);
          end
      end;
      fRedoList.Insert(0, s);
      MainForm.btnRedo.Enabled := True;
    end;
  end;
end;

procedure TBlockList.UndoSave(Cmd: string; Piece: TObject);
begin
  fUndoList.Insert(0, Cmd + ':' + TBlock(Piece).AsString);
  fRedoList.Clear;
  MainForm.btnUndo.Enabled := True;
  MainForm.btnRedo.Enabled := False;
end;

function TBlockList.RedoCount: integer;
begin
  Result := fRedoList.Count;
end;

function TBlockList.UndoCount: integer;
begin
  Result := fUndoList.Count;
end;

function TBlockList.LocateAsString(const s: string): integer;
var
  i                 : integer;
begin
  for i := 0 to Pred(Count) do begin
    if (TGeneralBlock(Items[i]).AsString = s) then begin
      Result := i;
      Exit;
    end
  end;
  Result := -1;
end;

procedure TBlockList.Add(const Piece: TObject);
var
  j                 : integer;
begin
  (**)
  { Faster, almost correct }
  j := Pred(Count);
  while (j >= 0) and Cover(Items[j], TBlock(Piece)) do
    Dec(j);
  if (j < Count) then
    inherited Insert(j + 1, Piece)
  else
    inherited Add(Piece);
  (**)
  (*
  { Super simple, correct, but slow }
    inherited Add(Piece);
    Oct := -1;  { Force sort }
  (**)
end;

procedure TBlockList.Concat(const Macro: TBlockList;
  const xx, yy, zz: integer);
var
  i                 : integer;
  mx, my, mz        : integer;
  b                 : TBlock;
  bn                : TBlock;
begin
  mx := 9999;
  my := 9999;
  mz := 9999;
  for i := 0 to Pred(Macro.Count) do begin
    with TBlock(Macro[i]) do begin
      if px < mx then mx := px;
      if py < my then my := py;
      if pz < mz then mz := pz;
    end
  end;
  for i := 0 to Pred(Macro.Count) do begin
    b := Macro[i];
    b.px := b.px - mx + xx;
    b.py := b.py - my + yy;
    b.pz := b.pz - mz + zz;
    bn := TBlock.CreateCopy(b);
    bn.Group := b.Group;
    Add(bn);
  end;
  Oct := -1;
end;

function TBlockList.Cover(const i, j: TBlock): Boolean;
{ Does i cover j, ie, should be drawn later?  }
begin
  with i do
    case Octant of
      0:
        Result := ((px + pl) <= (j.px)) or
          ((py) >= (j.py + j.pb)) or
          ((pz) >= (j.pz + j.ph));
      1:
        Result := ((px + pl) <= (j.px)) or
          ((py + pb) <= (j.py)) or
          ((pz) >= (j.pz + j.ph));
      2:
        Result := ((px) >= (j.px + j.pl)) or
          ((py + pb) <= (j.py)) or
          ((pz) >= (j.pz + j.ph));
      3:
        Result := ((px) >= (j.px + j.pl)) or
          ((py) >= (j.py + j.pb)) or
          ((pz) >= (j.pz + j.ph));
      4:
        Result := ((px + pl) <= (j.px)) or
          ((py) >= (j.py + j.pb)) or
          ((pz + ph) <= (j.pz));
      5:
        Result := ((px + pl) <= (j.px)) or
          ((py + pb) <= (j.py)) or
          ((pz + ph) <= (j.pz));
      6:
        Result := ((px) >= (j.px + j.pl)) or
          ((py + pb) <= (j.py)) or
          ((pz + ph) <= (j.pz));
      7:
        Result := ((px) >= (j.px + j.pl)) or
          ((py) >= (j.py + j.pb)) or
          ((pz + ph) <= (j.pz));
      else
        Result := False;
    end;
end;

{ Double-Bubble sort, slow, but (almost) correct }
(*
procedure TBlockList.SortForOctant(const AnOctant: integer);
var i, j: integer;
begin
  if AnOctant <> Oct then begin
    for i := 1 to Pred(Count) do begin
      for j := 1 to Pred(Count) do begin
        if Cover(Items[j-1], Items[j]) then begin
          Exchange(j-1, j);
        end
      end
    end;
    Oct := AnOctant;
  end;
//  CheckTopAndBottom;
end;
(**)

{ New sorting routine }
(*
procedure TBlockList.SortForOctant(const AnOctant: integer);
var
  l, r, i           : integer;
begin
  l := 0;
  r := Pred(Count);
  repeat
    for i := l + 1 to r do
    begin
      if Cover(Items[i - 1], Items[i]) then
      begin
        Exchange(i - 1, i);
      end
    end;
    Dec(r);
    for i := r downto l + 1 do
    begin
      if Cover(Items[i - 1], Items[i]) then
      begin
        Exchange(i - 1, i);
      end
    end;
    Inc(l);
  until l >= r;
  Oct := AnOctant;
end;
(**)

(**)
{ Tree-sorting, much faster, but not entirely correct! }
{ Perhaps the problem lies in Cover(...) ??? }
type
  PTreeNode = ^TTreeNode;
  TTreenode = record
    b: TBlock;
    l: PTreeNode;
    r: PTreeNode;
  end;

procedure TBlockList.SortForOctant(const AnOctant: integer);
var
  i                 : integer;
  tree              : PTreeNode;

  procedure AddNode(t: PTreeNode; b: TBlock);
  begin
    while True do begin
      if Cover(b, t.b) then begin
        if (t.r = nil) then begin
          New(t.r);
          t.r.l := nil;
          t.r.b := b;
          t.r.r := nil;
          break;
        end
        else begin
          t := t.r;
        end
      end
      else begin
        if (t.l = nil) then begin
          New(t.l);
          t.l.l := nil;
          t.l.b := b;
          t.l.r := nil;
          break;
        end
        else begin
          t := t.l;
        end
      end
    end
  end;

  procedure TraverseAdd(t: PTreeNode);
  begin
    if (t.l <> nil) then TraverseAdd(t.l);
    BlockList.Add(t.b);
    if (t.r <> nil) then TraverseAdd(t.r);
  end;

  procedure TraverseDel(t: PTreeNode);
  begin
    if (t.l <> nil) then TraverseDel(t.l);
    t.l := nil;
    t.b := nil;
    if (t.r <> nil) then TraverseDel(t.r);
    t.r := nil;
    Dispose(t);
  end;

begin
  if (BlockList.Count > 1) { and AnOctant <> Oct} then begin
    New(Tree);
    try
      tree.b := BlockList[0];
      tree.l := nil;
      tree.r := nil;
      for i := 1 to Pred(Count) do begin
        AddNode(Tree, TBlock(BlockList[i]));
      end;
    finally
      BlockList.ClearKeepUndo;
      TraverseAdd(Tree);
      TraverseDel(Tree);
    end;
    Oct := AnOctant;
  end;
  //  CheckTopAndBottom;
end;
(**)

procedure TBlockList.CheckTopAndBottom;
var
  i, j, c           : integer;
  mx, my, xx, xy    : integer;
  bi, bj            : TBlock;
begin
  if not MainForm.mHideStuds.Checked then
    Exit;
  for i := 0 to Pred(Count) do begin
    with TBlock(Items[i]) do begin
      TopCover := 0;
      BotCover := 0;
    end;
  end;
  for i := 0 to Pred(Pred(Count)) do begin
    Bi := TBlock(Items[i]);
    for j := 1 to Pred(Count) do begin
      Bj := TBlock(Items[j]);
      if (Bj.Transp) or (Bj.cl = clAqua) then
        continue;
      { Directly on top ?}
      if (Bj.pz = Bi.pz + Bi.ph) then begin
        { Overlapping ?}
        mx := Max(Bi.px, Bj.px);
        my := Max(Bi.py, Bj.py);
        xx := Min(Bi.px + Bi.pl, Bj.px + Bj.pl);
        xy := Min(Bi.py + Bi.pb, Bj.py + Bj.pb);
        if (xx > mx) and (xy > my) then begin
          c := (xx - mx) * (xy - my);
          Inc(Bi.TopCover, c);
          Inc(Bj.BotCover, c);
        end
      end
    end
  end;
end;

function TBlockList.FindBlockAtPos(const xx, yy, zz: integer): integer;
var
  i                 : integer;
begin
  Result := -1;
  for i := 0 to Pred(Count) do begin
    with TBlock(Items[i]) do begin
      if (px <= xx) and ((px + pl) >= xx) and
        (py <= yy) and ((py + pb) >= yy) and
        (pz <= zz) and ((pz + ph) >= zz) then begin
        Result := i;
        Exit;
      end;
    end;
  end;
end;

function TBlockList.FindBlockAtScreenPos(const xx, yy: integer): integer;
var
  i                 : integer;
begin
  Result := -1;
  for i := Pred(Count) downto 0 do begin
    with TBlock(Items[i]) do begin
      if Cutting.Dlg.BlockVisible(Items[i]) then
        if HitTest(xx, yy) then begin
          Result := i;
          Exit;
        end;
    end;
  end;
end;

procedure TBlockList.Draw(const Canvas: TDrawObject; DrawAlways: Boolean);
var
  i                 : integer;
  b                 : TBlock;
begin
  for i := 0 to Pred(Count) do begin
    b := TBlock(Items[i]);
    if Cutting.Dlg.BlockVisible(b) or DrawAlways then begin
      if Self.Step >= b.Step then
        b.Draw(Canvas);
    end
  end;
end;


procedure CalcStandardShades;
var
  t                 : TMyFloat;
begin
  t := sqrt(light.x * light.x + light.y * light.y + light.z * light.z);

  lfx := t * sqrt(PLX.A * PLX.A + PLX.B * PLX.B + PLX.C * PLX.C);
  if (lfx <> 0) then
    lfx := abs((light.x * PLX.A + light.y * PLX.B - light.z * PLX.C) / lfx / 4) + Threshold
  else
    lfx := Threshold;
  if lfx > 1.0 then
    lfx := 1.0;

  lfy := t * sqrt(PLY.A * PLY.A + PLY.B * PLY.B + PLY.C * PLY.C);
  if (lfy <> 0) then
    lfy := abs((light.x * PLY.A + light.y * PLY.B - light.z * PLY.C) / lfy / 4) + Threshold
  else
    lfy := Threshold;
  if lfy > 1.0 then
    lfy := 1.0;

  lfz := t * sqrt(PLZ.A * PLZ.A + PLZ.B * PLZ.B + PLZ.C * PLZ.C);
  if (lfz <> 0) then
    lfz := abs((light.x * PLZ.A + light.y * PLZ.B - light.z * PLZ.C) / lfz / 4) + Threshold
  else
    lfz := Threshold;
  if lfz > 1.0 then
    lfz := 1.0;

  lfsx := t * sqrt(PLSX.A * PLSX.A + PLSX.B * PLSX.B + PLSX.C * PLSX.C);
  if (lfsx <> 0) then
    lfsx := abs((light.x * PLSX.A + light.y * PLSX.B - light.z * PLSX.C) / lfsx / 4) + Threshold
  else
    lfsx := Threshold;
  if lfsx > 1.0 then
    lfsx := 1.0;

  lfsx1 := t * sqrt(PLSX1.A * PLSX1.A + PLSX1.B * PLSX1.B + PLSX1.C * PLSX1.C);
  if (lfsx1 <> 0) then
    lfsx1 := abs((light.x * PLSX1.A + light.y * PLSX1.B - light.z * PLSX1.C) / lfsx1 / 4) + Threshold
  else
    lfsx1 := Threshold;
  if lfsx1 > 1.0 then
    lfsx1 := 1.0;

  lfsy := t * sqrt(PLSY.A * PLSY.A + PLSY.B * PLSY.B + PLSY.C * PLSY.C);
  if (lfsy <> 0) then
    lfsy := abs((light.x * PLSY.A + light.y * PLSY.B - light.z * PLSY.C) / lfsy / 4) + Threshold
  else
    lfsy := Threshold;
  if lfsy > 1.0 then
    lfsy := 1.0;

  lfsy1 := t * sqrt(PLSY1.A * PLSY1.A + PLSY1.B * PLSY1.B + PLSY1.C * PLSY1.C);
  if (lfsy1 <> 0) then
    lfsy1 := abs((light.x * PLSY1.A + light.y * PLSY1.B - light.z * PLSY1.C) / lfsy1 / 4) + Threshold
  else
    lfsy1 := Threshold;
  if lfsy1 > 1.0 then
    lfsy1 := 1.0;

end;

procedure Shade(const c: TColor; const f: TMyFloat; var Res: TColor);
var
  RGB               : TColor;
  R, B, G           : integer;
begin
  RGB := ColorToRGB(c);
  B := Round(f * (RGB and $FF0000)) and $FF0000; {}
  R := Round(f * (RGB and $00FF00)) and $00FF00; {}
  G := Round(f * (RGB and $0000FF)) and $0000FF; {}
  Res := B or R or G;
end;

procedure ShadeEdge(const c1, c2: TColor; const f: TMyFloat; var Res: TColor);
var
  RGB               : TColor;
  R, B, G           : integer;
begin
  if (c1 = BlackColor) or (c1 = BlueColor) then
    RGB := ColorToRGB(c2)
  else
    RGB := ColorToRGB(c1);
  B := Round(f * (RGB and $FF0000)) and $FF0000; {}
  R := Round(f * (RGB and $00FF00)) and $00FF00; {}
  G := Round(f * (RGB and $0000FF)) and $0000FF; {}
  Res := B or R or G;
end;

procedure TBlockList.LoadFromFile(FileName: string);

  procedure ReadSettings(s: string);
  var
    i               : integer;
  begin
    i := 1;
    fSettingsRotXPos := NextInt(s, i, 0);
    fSettingsRotZPos := NextInt(s, i, 0);
    fSettingsScaleBarPos := NextInt(s, i, 0);
    fSettingsOffsetXPos := NextInt(s, i, 0);
    fSettingsOffsetYPos := NextInt(s, i, 0);
    fSettingsStuds := NextInt(s, i, 0);
    fSettingsVersion := NextInt(s, i, 0);
  end;

var
  sl                : TStringList;
  s                 : string;
  q                 : integer;
  i                 : integer;
  t                 : char;
  blk               : TBlock;
  PartFile          : string;
  PartName          : string;
  xMin, yMin, zMin  : integer;
  xMax, yMax, zMax  : integer;
begin
  xMin := 99999;
  yMin := 99999;
  zMin := 99999;
  xMax := -99999;
  yMax := -99999;
  zMax := -99999;
  sl := TStringList.Create;
  try
    sl.LoadFromFile(FileName);
    for q := 0 to Pred(sl.Count) do begin
      try
        t := sl[q][1];
        blk := nil;
        case t of
          'T': blk := TRoof.Create;
          'N': blk := TRoofTop.Create;
          'E': blk := TRoofTopEnd.Create;
          'C': blk := TRoofCorner.Create;
          'I': blk := TRoofTopCorner.Create;
          'R': blk := TRoundBlock.Create;
          'W': blk := TWindow.Create;
          'D': blk := TDoor.Create;
          'O': blk := TWheel.Create;
          'G': blk := TGeneralBlock.Create;
          '=': begin
              ReadSettings(Trim(Copy(sl[q], 2, 255)));
              continue;
            end
          else
            blk := TBlock.Create;
        end;
        if (t = ' ') then
          t := 'B';
        blk.t := t;
        s := Trim(Copy(sl[q], 2, 255));
        i := 1;
        with blk do begin
          case fSettingsVersion of
            0: begin
                px := NextInt(s, i, 0) * PFACTORXY;
                py := NextInt(s, i, 0) * PFACTORXY;
                pz := NextInt(s, i, 0) * PFACTORZ div 3;
                pl := NextInt(s, i, 0) * PFACTORXY;
                pb := NextInt(s, i, 0) * PFACTORXY;
                ph := NextInt(s, i, 0) * PFACTORZ div 3;
              end;
            1: begin
                px := NextInt(s, i, 0);
                py := NextInt(s, i, 0);
                pz := NextInt(s, i, 0);
                pl := NextInt(s, i, 0);
                pb := NextInt(s, i, 0);
                ph := NextInt(s, i, 0);
              end;
          end;
          cl := NextInt(s, i, 0);
          st := (NextInt(s, i, 0) <> 0);
          bp := (NextInt(s, i, 0) <> 0);
          Dir := NextInt(s, i, 0);
          Transp := (NextInt(s, i, 0) <> 0);
          if (t = 'T') and ((Odd(Dir) and (pl = (2 * PFACTORXY)) and (pb = (3 * PFACTORXY))) or
            (not Odd(Dir) and (pl = (3 * PFACTORXY)) and (pb = (2 * PFACTORXY)))) then
            LDrawPatch := 1;
          if (t = 'G') then begin
            PartFile := NextString(s, i, 'parts.dat');
            PartName := NextString(s, i, 'stdblock');
            (blk as TGeneralBlock).Initialize(PartFile, PartName);
          end;

          Group := NextInt(s, i, 0);
          NextGroup := Max(NextGroup, Group);
          PartID := NextString(s, i, '');
          Step := NextInt(s, i, 0);

          Add(blk);

          if not bp then begin
            xMin := Min(xMin, px);
            yMin := Min(yMin, py);
            zMin := Min(zMin, pz);
            xMax := Max(xMax, px + pl);
            yMax := Max(yMax, py + pb);
            zMax := Max(zMax, pz + ph);
          end
        end
      except
        on e: Exception do
          ShowMessage(e.Message);
      end
    end;
  finally
    sl.Free;
  end;
  bx := xMin;
  by := yMin;
  bz := zMin;
  bl := xMax - xMin;
  bb := yMax - yMin;
  bh := zMax - zMin;
  Oct := -2;
end;

initialization
  rx := 0;
  ry := 0;
  rz := 0;
  StudSwitcher := FastStud;
  DetailLevel := 0;

  calc_zplane(PLX, Point3D(0, 0, 0), Point3D(0, 0, 10), Point3D(10, 0, 10));
  calc_zplane(PLY, Point3D(0, 0, 0), Point3D(0, 10, 0), Point3D(0, 10, 10));
  calc_zplane(PLZ, Point3D(0, 0, 0), Point3D(10, 0, 0), Point3D(0, 10, 0));
  calc_zplane(PLSX, Point3D(0, 0, 4), Point3D(20, 20, 24), Point3D(40, 0, 4));
  calc_zplane(PLSX1, Point3D(0, 40, 4), Point3D(20, 20, 24), Point3D(40, 40, 4));
  calc_zplane(PLSY, Point3D(0, 0, 4), Point3D(0, 40, 4), Point3D(20, 20, 24));
  calc_zplane(PLSY1, Point3D(40, 0, 4), Point3D(40, 40, 4), Point3D(20, 20, 24));
  CalcStandardShades;
end.

